from __future__ import annotations as _

import collections.abc
import re
import threading
import time
import typing
import warnings

# [TODO] Implementation of repeat events
# [TODO] Fix a type error in CEventHandling.bind()
# [TODO] Support unbind for another widget's event


class CBoundTask:
    """A class to represent bound Task when a event is triggered."""

    def __init__(
        self,
        id_: str,
        target: typing.Callable | typing.Iterable,
        multithread: bool = False,
        _keep_at_clear: bool = False,
    ):
        """Each object is to represent a Task bound to the event.

        Example
        -------
        This is mostly for internal use of suzaku.
        .. code-block:: python
            class CEventHandling():
                def bind(self, ...):
                    ...
                    Task = CBoundTask(event_id, target, multithread, _keep_at_clear)
                    ...
        This shows where this class is used for storing Task properties in most cases.

        :param id_: The Task id of this Task
        :param target: A callable thing, what to do when this Task is executed
        :param multithread: If this Task should be executed in another thread (False by default)
        :param _keep_at_clear: If the Task should be kept when cleaning the event's binding
        """
        self.id: str = id_
        self.target: typing.Callable | typing.Iterable = target
        self.multithread: bool = multithread
        self.keep_at_clear: bool = _keep_at_clear


class CDelayTask(CBoundTask):
    """A class to represent delay Tasks"""

    def __init__(
        self,
        id_: str,
        target: typing.Callable | typing.Iterable,
        delay_: str,
        *args,
        **kwargs,
    ):
        """Inherited from CBoundTask, used to store Tasks bound to `delay` events.

        :param delay: Time to delay, in seconds, indicating how log to wait before the Task is
                      executed.
        :param (Other): See `CBoundTask.__init__()`
        """
        CBoundTask.__init__(self, id_, target, *args, **kwargs)  # For other things,same as
        # CBoundTask
        if delay_.endswith("s"):
            if delay_.endswith("ms"):
                delay_ = float(delay_[:-2]) / 1000
            else:
                delay_ = float(delay_[:-1])
        else:
            delay_ = float(delay_)
        self.target_time = float(time.time()) + delay_  # To store when to execute the Task


class CRepeatTask(CBoundTask):
    """A class to represent repeat Tasks"""

    def __init__(self, id_: str, target: typing.Callable, interval, *args, **kwargs):
        """Inherited from CBoundTask, used to store Tasks bound to `repeat` events.

        :param delay: Time to delay, in seconds, indicating how log to wait before the Task is
                      executed.
        """
        CBoundTask.__init__(self, id_, target, *args, **kwargs)  # For other things,same as
        # CBoundTask
        self.target_time = float(time.time()) + interval  # To store when to execute the Task for
        # the next time, will be accumulated after
        # execution of the Task
        self.interval = interval  # Interval of the Task


class CEventHandling:
    """A class containing event handling abilities.

    This class should be inherited by other classes with such abilities.

    Events should be represented in the form of `event_type` or `event_type[args]`. e.g. `delay` or
    `delay[500]`
    """

    # fmt: off
    EVENT_TYPES: list[str] = [
        "resize", "move", 
        "configure", "update", "redraw", 
        "mouse_move", "mouse_enter", "mouse_leave", "mouse_press", "mouse_release", "click", "double_click",
        "focus_gain", "focus_loss", 
        "key_press", "key_release", "key_repeat", "char", 
        "delay", "repeat", # This row shows special event type(s)
    ]
    # fmt: on
    multithread_Tasks: list[tuple[CBoundTask, CEvent]] = []
    WORKING_THREAD: threading.Thread
    instance_count = 0

    @classmethod
    def _working_thread_loop(cls):
        while True:
            try:
                CEventHandling._execute_Task(
                    cls.multithread_Tasks[0][0], cls.multithread_Tasks[0][1]
                )
                # For line above: [0][0] is Task object, [0][1] is CEvent object
            except Exception as e:
                warnings.warn(
                    RuntimeWarning(
                        "Error in multithread suzaku-event-bound Task "
                        f"with ID {cls.multithread_Tasks[0][0].id}, "
                        f'detailed error info: "{str(e)}". '
                        "The working thread will head to the next Task "
                        "(Cipping current causing the error)."
                    )
                )
            cls.multithread_Tasks.pop(0)  # Executed Tasks should be removed, anyway

            if not cls.multithread_Tasks:
                time.sleep(0.01)  # Avoid CPU draining while no Tasks avail
                continue

    @staticmethod
    def _execute_Task(Task: CBoundTask, event_obj: CEvent) -> None:
        """To execute the binded Task directly, regardless its props, mainly for internal use."""
        match Task.target:
            case _ if callable(Task.target):
                Task.target(event_obj)
            case _ if isinstance(Task.target, collections.abc.Iterable):
                for Task_step in Task.target:
                    Task_step(event_obj)
            case _:
                raise ValueError(
                    "Error type for suzaku Task target! Excepted callable or "
                    f"iterable but received {type(Task.target)}"
                )

    def __init__(self):
        """A class containing event handling abilities.

        Example
        -------
        This is mostly for internal use of Charmy.
        .. code-block:: python
            class CWidget(CEventHandling, ...):
                def __init__(self):
                    super().__init__(self)
            ...
        This shows subclassing CEventHandling to let CWidget gain the ability of handling events.
        """
        self.latest_event: CEvent = CEvent(widget=None, event_type="NO_EVENT")
        self.Tasks: dict[str, list[CBoundTask]] = {}
        # Make a initial ID here as it will be needed anyway even if the object does not have an ID.
        self.id = f"{self.__class__.__name__}{self.__class__.instance_count}"
        ## Initialize Tasks list
        for event_type in self.__class__.EVENT_TYPES:
            self.Tasks[event_type] = []
        ## Accumulate instance count
        self.__class__.instance_count += 1
        # Event binds
        self.bind("update", self._check_delay_events, _keep_at_clear=True)  # Delay checking loop

    def parse_event_type_str(self, event_type_str) -> dict:
        """This function parses event type string.

        :param event_type_str: The event type string to be parsed
        :returns: json, parsed event type
        """
        if not re.match(".*\\[.*\\]", event_type_str):  # NOQA
            return {"type": event_type_str, "params": []}
        event_type = re.findall("^(.*?)\\[", event_type_str)[0]
        params_raw = re.findall("\\[(.*?)\\]$", event_type_str)[0]  # NOQA
        params = params_raw.split(",")
        if len(params) == 1:
            if params[0].strip() == "":
                params = []
        return {"type": event_type, "params": params}

    def execute_Task(self, Task: CBoundTask, event_obj: CEvent | None = None):
        """To execute a Task

        Example
        -------
        .. code-block:: python
            my_Task = CWidget.bind("delay[5]", lambda: print("Hello Suzaku"))
            CWidget.execute_Task(my_Task)
        """
        if event_obj is None:
            event_obj = CEvent()
        assert event_obj is not None
        if event_obj.widget is None:
            event_obj.widget = self
        if not Task.multithread:
            # If not multiTask, execute directly
            CEventHandling._execute_Task(Task, event_obj)
            # If is a delay event, it should be removed right after execution
            if isinstance(Task, CDelayTask):
                self.unbind(Task)
        else:
            # Otherwise add to multithread Tasks list and let the working thread to deal with it
            # If is a delay Task, should add some code to let it unbind itself, here is a way,
            # which is absolutely not perfect, though works, to implement this mechanism, by
            # overriding its target with a modified version
            def self_destruct_template(Task, event_obj):
                CEventHandling._execute_Task(Task, event_obj)
                self.unbind(Task)

            if isinstance(Task, CDelayTask):
                Task.target = lambda event_obj: self_destruct_template(Task, event_obj)
            CEventHandling.multithread_Tasks.append((Task, event_obj))

    def trigger(self, event_type: str, event_obj: CEvent | None = None) -> None:
        """To trigger a type of event

        Example
        -------
        .. code-block:: python
            class CWidget(CEventHandling, ...):
                ...

            my_widget = CWidget()
            my_widget.trigger("mouse_press")
        This shows triggering a `mouse_press` event in a `CWidget`, which inherited `CEventHandling` so has the
        ability to handle events.

        :param event_type: The type of event to trigger
        """
        # Parse event type string
        parsed_event_type = self.parse_event_type_str(event_type)
        # Create a default CEvent object if not specified
        if event_obj is None:
            event_obj = CEvent(widget=self, event_type=tuple(parsed_event_type.keys())[0])
        # Add the event to event lists (the widget itself and the global list)
        self.latest_event = event_obj
        CEvent.latest = event_obj
        # Find targets
        targets = [parsed_event_type["type"]]
        if not parsed_event_type["params"]:
            targets.append(parsed_event_type["type"] + "[*]")
        else:
            targets.append(event_type)
        # if parsed_event_type["params"][0] in ["", "*"]: # If match all
        #     targets.append(parsed_event_type["type"])
        #     targets.append(parsed_event_type["type"] + "[*]")
        for target in targets:
            if target in self.Tasks:
                for Task in self.Tasks[target]:
                    # To execute all Tasks bound under this event
                    self.execute_Task(Task, event_obj)

    def bind(
        self,
        event_type: str,
        target: typing.Callable | typing.Iterable,
        multithread: bool = False,
        _keep_at_clear: bool = False,
    ) -> CBoundTask | bool:
        """To bind a Task to the object when a specific type of event is triggered.

        Example
        -------
        .. code-block
            my_button = CButton(...).pack()
            press_down_event = my_button.bind("mouse_press", lambda _: print("Hello world!"))
        This shows binding a hello world to the button when it's press.

        :param event_type: The type of event to be bound to
        :param target: A (list of) callable thing, what to do when this Task is executed
        :param multithread: If this Task should be executed in another thread (False by default)
        :param _keep_at_clear: If the Task should be kept when cleaning the event's binding
        :return: CBoundTask that is bound to the Task if success, otherwise False
        """
        parsed_event_type = self.parse_event_type_str(event_type)
        if parsed_event_type["type"] not in self.__class__.EVENT_TYPES:
            # warnings.warn(f"Event type {event_type} is not present in {self.__class__.__name__}, "
            #                "so the Task cannot be bound as expected.")
            # return False
            self.EVENT_TYPES.append(event_type)
        if event_type not in self.Tasks:
            self.Tasks[event_type] = []
        Task_id = f"{self.id}.{event_type}.{len(self.Tasks[event_type])}"
        # e.g. CButton114.focus_gain.514 / CEventHandling114.focus_gain.514
        match parsed_event_type["type"]:
            case "delay":
                Task = CDelayTask(
                    Task_id,
                    target,  # I will fix this type error later (ignore is ur type check is off)
                    parsed_event_type["params"][0],
                    multithread,
                    _keep_at_clear,
                )
            case "repeat":
                raise NotImplementedError("repeat events is not implemented yet!")
            case _:  # All normal event types
                Task = CBoundTask(Task_id, target, multithread, _keep_at_clear)
        self.Tasks[event_type].append(Task)
        return Task

    def clear_bind(self, event_type: str) -> bool:
        """To clear clear Tasks binded to a spcific event or widget

        Example
        -------
        .. code-block:: python
            my_widget = CWidget(...)
            my_widget.clear_bind("click")
        This shows clearing Tasks bound to `click` event on `my_widget`.

        .. code-block:: python
            my_widget = CWidget(...)
            my_widget.clear_bind("*")
        This shows clearing all events bound to any event on `my_widget`.

        :param event_type: Type of event to clear binds, `*` for all
        :return: Boolean, whether success or not
        """
        if event_type == "*":  # Clear all Tasks of this object
            return not False in [self.clear_bind(this_type) for this_type in self.Tasks]
        else:  # In other cases, this must be an specific event type
            if event_type in self.Tasks:  # If type given existed and include some Tasks
                for Task in self.Tasks[event_type]:
                    if not Task.keep_at_clear:  # Cip any keep_at_clear Tasks
                        self.unbind(Task)
                return True
            else:
                return False

    def _check_delay_events(self, _=None) -> None:
        """To check and execute delay events.

        Example
        -------
        Mostly used by CWidget.update(), which is internal use.

        :param _: To accept an event object if is given, will be ignored
        """
        # print("Checking delayed events...")
        for binded_event_type in tuple(self.Tasks):
            if self.parse_event_type_str(binded_event_type)["type"] == "delay":
                for Task in self.Tasks[binded_event_type]:
                    if isinstance(Task, CDelayTask):
                        if float(time.time()) >= Task.target_time:
                            # print(f"Current time is later than target time of {Task.id}, "
                            #        "execute the Task.")
                            self.execute_Task(Task)


# Initialize working thread
CEventHandling.WORKING_THREAD = threading.Thread(target=CEventHandling._working_thread_loop)


# @daTasklass
class CEvent:
    """Used to represent an event."""

    latest: CEvent

    def __init__(
        self,
        widget: CEventHandling | None = None,
        event_type: str = "[Unspecified]",
        **kwargs,
    ):
        """This class is used to represent events.

        Some properties owned by all types of events are stored as attributes, such as widget and type.
        Others are stored as items, which can be accessed or manipulated just like dict, e.g.
        `CEvent["x"]` for get and `CEvent["y"] = 16` for set.

        Example
        -------
        Included in description.

        :param widget: The widget of the event, None by default
        :param event_type: Type of the event, in string, `"[Unspecified]"` by default
        :param **kwargs: Other properties of the event, will be added as items
        """
        self.event_type: str = event_type  # Type of event
        self.widget: typing.Optional[typing.Any] = widget  # Relating widget
        self.window_base: typing.Optional[typing.Any] = None  # WindowBase of the current window
        self.window: typing.Optional[typing.Any] = None  # Current window
        self.event_data: dict = {}
        # Not all properties above will be used
        # Update stuff from args into attributes
        for prop in kwargs.keys():
            if prop not in ["widget", "event_type"]:
                self[prop] = kwargs[prop]

    def __setitem__(self, key: str, value: typing.Any):
        self.event_data[key] = value

    def __getitem__(self, key: str) -> typing.Any:
        if key in self.event_data:
            return self.event_data[key]
        else:
            return None  # If no such item avail, returns None


CEvent.latest = CEvent(widget=None, event_type="NO_EVENT")



import typing
import warnings

from ..const import ID, MANAGER_ID
from ..object import CharmyObject
from ..rect import Rect
from .cmm import CharmyManager


class CanvasBase(CharmyObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Auto find CharmyManager Object
        manager = self.get_obj(MANAGER_ID)
        if manager is None:
            raise ValueError("Not found main CharmyManager")
        self.manager: CharmyManager = manager
        self.new(
            "drawing.framework",
            self._get_drawing_framework(),
            get_func=self._get_drawing_framework,
        )
        self.new(
            "drawing.framework.class",
            self._get_drawing_framework_class(),
            get_func=self._get_drawing_framework_class,
        )
        # element config
        self.elements = []  # e.g. [{"type": "rect", "id": "element0", "radius": 12}]
        self.draw_type_map: dict[str, typing.Callable[[dict, dict], None]] = {
            "rect": self.draw_rect,
        }  # 元素绘制映射表
        self.new("color_object", None)
        self._last_count = 0

    def draw(self, canvas):
        """Draw the widget"""
        for element in self.elements:
            if element["type"] in self.draw_type_map:
                self.draw_type_map[element["type"]](canvas, element)
            else:
                warnings.warn(f"Warning: Unknown element type {element['type']}")

    def draw_rect(self, canvas, element: dict):
        self["drawing.framework.class"].draw_rect(
            canvas,
            rect=element.get("rect"),
            radius=element.get("radius", 0),
            bg=element.get("bg", None),
        )

    def _template(self, type_: str, rect: Rect, id_: ID = ID.AUTO, **kwargs):
        """Create a new element.

        Args:
            type_ (str): Element type
            rect (Rect): Element position and size
            id_ (ID, optional): Element ID. Defaults to ID.AUTO.

        Returns:
            dict: Element config dict
        """
        if id_ == ID.AUTO:
            id_ = f"element{self._last_count}"

        self._last_count += 1

        return {"type": type_, "rect": rect, "id": id_, **kwargs}

    # region Element config

    def add_element(self, type_: str, rect: Rect, **kwargs):
        self.elements.append(self._template(type_, rect, **kwargs))

    def insert_element(self, index: int, type_: str, rect: Rect, **kwargs):
        self.elements.insert(index, self._template(type_, rect, **kwargs))

    def remove_element(self, index: int):
        self.elements.remove(index)

    def find_element(self, id_: ID) -> dict | None:
        for element in self.elements:
            if element["id"] == id_:
                return element
        return None

    get_element = find_element

    def _get_drawing_framework(self):
        return self.manager["drawing.framework"]

    def _get_drawing_framework_class(self):
        return self.manager.get("drawing.framework.class")



from .object import CharmyObject


class Rect(CharmyObject):
    """Rect is a class to store the position and size of a rectangle.

    Args:
        **kwargs: The keyword arguments to initialize the rectangle.
            The keyword arguments are:
                x (int): The x coordinate of the top-left corner of the rectangle.
                y (int): The y coordinate of the top-left corner of the rectangle.
                width (int): The width of the rectangle.
                height (int): The height of the rectangle.
                left (int): The left coordinate of the top-left corner of the rectangle.
                top (int): The top coordinate of the top-left corner of the rectangle.
                right (int): The right coordinate of the top-left corner of the rectangle.
                bottom (int): The bottom coordinate of the top-left corner of the rectangle.
    """

    def __init__(self, **kwargs):
        super().__init__()
        # 基础四要素
        self.new("base_x", 0)
        self.new("base_y", 0)
        self.new("base_width", 0)
        self.new("base_height", 0)

        #
        self.new("top", 0, set_func=self._set_top, get_func=self._get_top)
        self.new("left", 0, set_func=self._set_left, get_func=self._get_left)
        self.new("bottom", 0, set_func=self._set_bottom, get_func=self._get_bottom)
        self.new("right", 0, set_func=self._set_right, get_func=self._get_right)
        self.new("x", 0, set_func=self._set_left, get_func=self._get_left)  # = left
        self.new("y", 0, set_func=self._set_top, get_func=self._get_top)  # = top
        self.new("width", 0, set_func=self._set_width, get_func=self._get_width)
        self.new("height", 0, set_func=self._set_height, get_func=self._get_height)
        self.__call__(**kwargs)

    def __call__(self, **kwargs):
        self.set("x", kwargs.get("x", 0))
        self.set("y", kwargs.get("y", 0))
        self.set("width", kwargs.get("width", 0))
        self.set("height", kwargs.get("height", 0))
        #############################################################
        self.set("left", kwargs.get("left", 0))
        self.set("top", kwargs.get("top", 0))
        self.set("right", kwargs.get("right", 0))
        self.set("bottom", kwargs.get("bottom", 0))

    def __str__(self):
        """Return position in string."""
        return f"Rect({self.get('left')}, {self.get('top')}, {self.get('width')}, {self.get('height')})"

    def _set_left(self, x):
        self.set("base_x", x)

    def _get_left(self):
        return self.get("base_x")

    def _set_top(self, y):
        self.set("base_y", y)

    def _get_top(self):
        return self.get("base_y")

    def _set_right(self, x):
        self.set("base_width", x - self.get("base_x"))

    def _get_right(self):
        return self.get("base_x") + self.get("base_width")

    def _set_bottom(self, y):
        self.set("base_height", y - self.get("base_y"))

    def _get_bottom(self):
        return self.get("base_y") + self.get("base_height")

    def _set_width(self, w):
        self.set("base_width", w)

    def _get_width(self):
        return self.get("base_width")

    def _set_height(self, h):
        self.set("base_height", h)

    def _get_height(self):
        return self.get("base_height")



"""Widgets for Charmy"""

from .cmm import cquit, manager, mainloop
from .container import Container, auto_find_parent
from .widget import Widget
from .window import Window
from .windowbase import WindowBase



from .canvas import CanvasBase
from .container import Container, auto_find_parent


@auto_find_parent
class Widget(CanvasBase):
    """Basic widget class."""

    def __init__(self, parent: Container | None = None):
        super().__init__()

        if parent is None:
            parent = self.find("window0", default=None)

        self.new("parent", parent)

        # Automatically add to parent's children list if parent exists
        if parent is not None:
            parent.add_child(self)



from .object import CharmyObject


class Size(CharmyObject):
    """Size is a class to store size.

    To reset the size, use the __call__ method.

    Example:
        >>> size = Size(100, 200)
        >>> print(size)  # CSize(100, 200)
        >>> size(300, 400)  # or size.__call__(300, 400)
        >>> print(size)  # CSize(300, 400)
    ```
    """

    def __init__(self, width=0, height=0):
        super().__init__()
        self.new("width", width)
        self.new("height", height)

    def __call__(self, width: int | float | None = None, height: int | float | None = None):
        if width:
            self.set("width", width)
        if height:
            self.set("height", height)

    def __str__(self):
        return f"Size({self.get('width')}, {self.get('height')})"



import functools
import threading
import typing

from ..object import CharmyObject


class Container(CharmyObject):
    """CContainer is a class to store child objects"""

    _local = threading.local()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.new("children", [])

    # region Context

    @classmethod
    def _get_context_stack(cls) -> typing.List["Container"]:
        """Get the context stack of the current thread"""
        if not hasattr(cls._local, "context_stack"):
            cls._local.context_stack = []
        return cls._local.context_stack

    @classmethod
    def get_context(cls) -> typing.Optional["Container"]:
        """Get the current context container"""
        stack = cls._get_context_stack()
        return stack[-1] if stack else None

    def __enter__(self) -> "Container":
        """Enter the context"""
        stack = self._get_context_stack()
        stack.append(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the context"""
        stack = self._get_context_stack()
        if stack and stack[-1] is self:
            stack.pop()
        return False  # 不抑制异常

    def add_child(self, child: "CharmyObject"):
        """Add a child object"""
        if child not in self["children"]:
            self["children"].append(child)

    def draw_children(self, canvas):
        """Draw the container and its children"""
        for child in self["children"]:
            if hasattr(child, "draw"):
                child.draw(canvas)

    # endregion


def auto_find_parent(widget_class: typing.Callable) -> typing.Callable:
    """Add a decorator to automatically inject the parent container to the widget constructor"""

    # Save the original constructor
    original_init = widget_class.__init__  # NOQA

    @functools.wraps(original_init)
    def new_init(self, *args, **kwargs):
        # Check if parent is specified in keyword arguments
        parent_specified = False

        # Check if parent is specified in keyword arguments
        if "parent" in kwargs:
            parent_specified = True
        # Check if parent is specified in positional arguments
        elif len(args) >= 2:
            parent_specified = True

        # If parent is not specified, try to get it from context
        if not parent_specified:
            parent = Container.get_context()
            if parent is not None:
                kwargs["parent"] = parent

        # Call the original constructor
        original_init(self, *args, **kwargs)

    # Replace the constructor
    widget_class.__init__ = new_init
    return widget_class



"""Basic object class."""

import typing
import weakref

from .const import ID


class InstanceCounterMeta(type):
    """
    InstanceCounterMeta
    """

    def __init__(cls, name, bases, attrs):
        super().__init__(name, bases, attrs)
        cls._instances = weakref.WeakSet()

    def __call__(cls, *args, **kwargs):
        instance = super().__call__(*args, **kwargs)

        if type(instance) is cls:
            cls._instances.add(instance)

        return instance


class CharmyObject(metaclass=InstanceCounterMeta):
    """CharmyObject is this project's basic class, with features to set and get attribute."""

    objects: typing.Dict[str, typing.Any] = {}  # find by ID {1: OBJ1, 2: OBJ2}
    objects_sorted: typing.Dict[str, typing.Any] = (
        {}
    )  # find by class name {OBJ1: {1: OBJECT1, 2: OBJECT2}}

    def __init__(self, id_: ID | str = ID.AUTO):
        self._attributes: dict[str, typing.Any | list] = {}  # NOQA: Expected to be user-modifiable.
        # self._attributes -> {key: value, key2: ["@custom", value, set_func, get_func]}
        # self._attributes[key] -> ["@custom", value, set_func, get_func] | value

        if id_ == ID.AUTO:
            _prefix = self.class_name
            id_ = _prefix + str(self.instance_count)
        if id_ in self.objects:
            raise KeyError(id_)
        if id_ != ID.NONE:
            self.objects[id_] = self
            self.id = id_

            if self.class_name not in self.objects_sorted:
                self.objects_sorted[self.class_name] = {self.id: self}
            else:
                self.objects_sorted[self.class_name][self.id] = self

    @property
    def class_name(self) -> str:
        """Return the lowered class name"""
        return self.__class__.__name__.lower()

    @property
    def instances(self):
        """Return all the class instances"""
        return self.objects_sorted[self.class_name]

    @property
    def instance_count(self):
        """Return the class instance count"""
        return len(self._instances)

    def new(
        self,
        key: str,
        default=None,
        *,
        set_func: typing.Callable | None = None,
        get_func: typing.Callable | None = None,
    ) -> typing.Self:
        """New attribute
        [Suggested: add `is_`~ as prefix if value is in type bool]

        Example
        -------
        .. code-block:: python

            from charmy import *

            obj = CharmyObject()
            obj.new("name", "obj1")
            print(obj.get("name"))  # OUTPUT: obj1
            obj.set("name", "obj2")
            print(obj.get("name"))  # OUTPUT: obj2


            def set_func(value):
                print("User set value:", value)
                return "Hello, " + value


            def get_func():
                return f"User get value: {obj.get('title', skip=True)}"


            obj.new("title", "default title", set_func=set_func, get_func=get_func)
            print(obj.get("title"))  # OUTPUT: default title
            obj.set("title", "A")  # OUTPUT: User set value: A
            print(obj.get("title"))  # User get value: Hello, A


        Args:
            key (str): attribute name
            default (typing.Any): default value
            set_func (typing.Callable): function to set
            get_func (typing.Callable): function to get

        Return:
            Self
        """
        if set_func or get_func:
            self._attributes[key] = ["@custom", default, set_func, get_func]
        else:
            self._attributes[key] = default
        return self

    def set(self, key: str, value, *, skip: bool = False) -> typing.Self:
        """Set attribute.

        Args:
            key (str): attribute name
            value (typing.Any): attribute value
            skip (bool): skip custom set_func function

        Return:
            self
        """
        if key not in self._attributes:
            raise KeyError(key)
        if isinstance(self._attributes[key], list):
            if self._attributes[key][0] == "@custom":
                if self._attributes[key][2]:
                    if not skip:
                        _return = self._attributes[key][2](value)
                    else:
                        _return = None
                    if _return:
                        self._attributes[key][1] = _return
                    else:
                        self._attributes[key][1] = value
                else:
                    self._attributes[key][1] = value
                return self
        self._attributes[key] = value
        return self

    def get(self, key: str, default=None, *, skip: bool = False) -> typing.Any:
        """Get attribute by key.

        Args:
            key (str): attribute name
            skip (bool): skip custom get_func function

        Return:
            the value corresponding to the key
        """

        # check is an available key
        if key not in self._attributes:
            return default

        # check is a custom attribute
        if isinstance(self._attributes[key], list) and len(self._attributes[key]) > 0:
            if self._attributes[key][0] == "@custom":
                if self._attributes[key][3]:
                    if not skip:
                        _return = self._attributes[key][3]()
                    else:
                        _return = None

                    if _return:
                        return _return
                    else:
                        return self._attributes[key][1]
                else:
                    return self._attributes[key][1]
        return self._attributes[key]

    def get_obj(self, _id: str, default=None) -> typing.Any | None:
        """Get registered object by id. (If not found, return default)"""
        try:
            return self.objects[_id]
        except KeyError:
            return default

    find = get_obj

    def configure(self, **kwargs):
        """High level set attributes by keyword arguments."""
        for key, value in kwargs.items():
            self.set(key, value)

    config = configure

    def cget(self, key: str) -> typing.Any:
        """Low level get attribute by key"""
        return self.get(key)

    # region Item configuration

    def __setitem__(self, key: str, value: typing.Any):
        """Set attribute by key: obj["key"] = value"""
        self.set(key, value)

    def __getitem__(self, key: str) -> typing.Any:
        """Get attribute by key: obj["key"]"""
        return self.get(key)

    def __delitem__(self, key: str):
        """Delete attribute by key: obj["key"]"""
        del self._attributes[key]

    def __contains__(self, item: str) -> bool:
        return item in self._attributes

    # endregion

    @property
    def attributes(self) -> typing.Dict[str, typing.Any]:
        """Return all attributes"""
        return self._attributes

    def __str__(self) -> str:
        """Return all attributes in string"""
        return str(self._attributes)



import inspect
import json
import os
import os.path
import pathlib
import typing
import warnings

from ..object import CharmyObject


class Theme(CharmyObject):
    """Theme class for CWindow and CWidgets.

    Example
    -------
    .. code-block:: python
        my_theme = CTheme({<Some styles>})
        my_sub_theme = CTheme(parent="default.light")
        my_external_theme = CTheme().load_from_file("./path/to/a/theme.json")
    This shows examples of creating themes, either from a JSON, a parent theme or a file.

    .. code-block:: python
        all_themes = CTheme.LOADED_THEMES
        INTERNAL_THEME = CTheme.INTERNAL_THEME
        DEFAULT_THEME = CTheme.DEFAULT_THEME
    This shows getting all loaded themes, internal themes, and the default theme.

    .. code-block:: python
        default_light_theme = CTheme.find_loaded_theme("default.light")
        if CTheme.validate_theme_existed("default.light"):
            print("Default light theme exists!")
    This shows finding a theme and checking if it exists

    重置后我摸改了一些地方，但逻辑、功能应是大差不差的，某些方法名、变量名我改的易读了点
    """

    LOADED_THEMES: list["Theme"] = []
    INTERNAL_THEME_DIR = pathlib.Path(__file__).parent.parent / "resources" / "themes"
    INTERNAL_THEME: dict[str, "Theme"] = {}
    DEFAULT_THEME: "Theme"
    DEFAULT_THEME_FILENAME: str = "light"
    EXPECTED_DATA_TYPE = {
        "styles": dict,
        "color_palette": dict,
        "name": str,
        "friendly_name": str,
        "base": str,
    }

    def __init__(
        self, styles: dict | None = None, parent: typing.Union["Theme", None] = None, *kwargs
    ) -> None:
        """Theme for CWindow and CWidgets.

        Example
        -------
        .. code-block:: python
            my_theme = CTheme({<Some styles>})
            my_sub_theme = CTheme(parent="default.light")
            my_external_theme = CTheme().load_from_file("./path/to/a/theme.json")
        This shows examples of creating themes, either from a JSON, a parent theme or a file.

        :param styles: Styles of the theme
        :param parent: Parent theme
        """
        super().__init__(**kwargs)  # NOQA

        self["name"]: str = f"untitled.{len(Theme.LOADED_THEMES) + 1}"
        self["friendly_name"] = f"Untitled theme {len(Theme.LOADED_THEMES) + 1}"
        # NOQA: friendly_name感觉有点多余? ——Little White Cloud
        # NOQA: Keep it 4 now currently. ——rgzz666
        self.parent: typing.Union["Theme", None] = parent
        self.children = []
        self["is_special"]: bool = False

        if styles is None:
            self["styles"]: dict = Theme.DEFAULT_THEME["styles"]
        else:
            self["styles"]: dict = styles
        self["color_palette"] = {}

        Theme.LOADED_THEMES.append(self)
        return

    @classmethod
    def find_loaded_theme(cls, theme_name: str) -> "Theme | typing.Literal[False]":
        """Search for a loaded theme by name, returns the CTheme object if found, or False if not.

        Example
        -------
        .. code-block:: python
            DEFAULT_THEME = CTheme.find_loaded_theme("default.light")
        This returns the CTheme object of the default theme to `DEFAULT_THEME`.

        :param theme_name: Name of the theme to load
        :return: The CTheme object if found, otherwise False
        """
        for theme in cls.LOADED_THEMES:
            if theme["name"] == theme_name:
                return theme
        return False

    @classmethod
    def validate_theme_existed(cls, theme_name: str) -> bool:
        """Validate if the theme with given name existed and loaded.

        Example
        -------
        .. code-block:: python
            CTheme.validate_theme_existed("default.light")
        This returns if the theme `default.light` is loaded.

        :param theme_name: Name of the theme to validate
        :return: If the theme loaded
        """
        return Theme.find_loaded_theme(theme_name) != False  # ☝🤓

    # region Load theme

    @classmethod
    def _load_internal_theme(cls):
        """Load internal themes. Should be run once at import, see the end of this file."""
        # Load default (ROOT) theme
        Theme.DEFAULT_THEME = typing.cast(
            "Theme",
            Theme({}).load_from_file(
                Theme.INTERNAL_THEME_DIR / f"{Theme.DEFAULT_THEME_FILENAME}.json"
            ),
        )

        # Load other internal themes
        for file in os.listdir(Theme.INTERNAL_THEME_DIR):
            if file == f"{Theme.DEFAULT_THEME_FILENAME}.json":
                # For default theme, no need to reload it
                Theme.INTERNAL_THEME[Theme.DEFAULT_THEME["name"]] = Theme.DEFAULT_THEME
                continue
            _ = Theme({}).load_from_file(Theme.INTERNAL_THEME_DIR / file)
            Theme.INTERNAL_THEME[_["name"]] = _

    def load_from_file(self, file_path: str | pathlib.Path) -> "Theme | typing.Literal[False]":
        """Load styles to theme from a file.

        Example
        -------
        .. code-block:: python
            my_theme = CTheme().load_from_file("./path/to/a/theme.json")
            my_theme.load_from_file("./path/to/another/theme.json")

        This shows loading a theme to `my_theme` from the theme file at `./path/to/a/theme.json`,
        and change it to theme from `./path/to/another/theme.json` later.

        :param file_path: Path to the theme file
        :return self: The CTheme itself
        """
        # Change path string into pathlib Path
        if type(file_path) is str:
            file_path = pathlib.Path(file_path)
        # Get path where lies codes calling this function (to support relative path)
        if not file_path.is_absolute():
            frame = inspect.currentframe()
            outer_frame = inspect.getouterframes(frame)[1]
            caller_file = pathlib.Path(outer_frame.filename).parent
            file_path = (caller_file / file_path).resolve()
        # We need a file to load from file \o/ \o/ \o/
        with open(file_path, mode="r", encoding="utf-8") as f:
            style_raw = f.read()

            theme_data = json.loads(style_raw)
            if (search_result := Theme.find_loaded_theme(theme_data["name"])) != False:
                # If name already occupied, meaning the theme might already be loaded
                # (or just simply has an occupied name)
                warnings.warn(
                    f"Theme <{theme_data['name']}> already loaded or existed.",
                    RuntimeWarning,
                )
                return search_result

        return self.load_from_json(theme_data)

    def load_from_json(self, theme_data: dict) -> "Theme":
        """Load all data (including metadata) to the theme.

        Example
        -------
        .. code-block:: python
            my_theme = CTheme().load_from_json({<Some JSON theme data>})
            my_theme.load_from_json({<Some JSON theme data>})
        This shows loading a theme to `my_theme` from JSON data, and change it to theme from
        another JSON later.

        :param theme_data: dict that contains the theme data
        :return self: The CTheme itself
        """
        # Type check
        for item in self.EXPECTED_DATA_TYPE.keys():
            if type(theme_data[item]) != self.EXPECTED_DATA_TYPE[item]:
                theme_name = (
                    theme_data["name"] if type(theme_data["name"]) is str else "(Type error)"
                )
                warnings.warn(
                    f"Error data type of <{item}> in theme data that is about to be loaded. "
                    f"Expected {self.EXPECTED_DATA_TYPE[item]} but got {type(item)}. The json data with "
                    f"theme named <{theme_name}> will not be loaded to the theme <{self['name']}>",
                    ResourceWarning,
                )
                return self
        # Load data
        self["styles"] = theme_data["styles"].copy()
        self["color_palette"] = theme_data["color_palette"].copy()
        # Load Metadata
        self.rename(theme_data["name"], theme_data["friendly_name"])
        self.set_parent(theme_data["base"])

        return self

    # endregion



import importlib
import typing
import warnings
from os import environ

from ..const import MANAGER_ID, Backends
from ..event import WorkingThread
from ..framework import drawing_framework_map, window_framework_map
from ..object import CharmyObject


class CharmyManager(CharmyObject):
    """The manager of the charmy windows.

    Args:
        ui.framework: What UI Framework will be used.
        ui.is_vsync: Is vsync enabled.
        ui.samples: UI Samples.
    """

    def __init__(
        self,
        ui: Backends = Backends.GLFW,
        drawing: Backends = Backends.SKIA,
        backend: Backends = Backends.OPENGL,
        vsync: bool = True,
        samples: int = 4,
        **kwargs,
    ):
        super().__init__(**kwargs)

        if self.instance_count >= 1:
            warnings.warn("There should be only one instance of CApp.")

        self.new("event.thread", WorkingThread())

        self.new("ui.framework", ui)
        self.new("ui.framework.class", window_framework_map[ui]())  # NOQA
        self.new("ui.is_vsync", vsync)
        self.new("ui.samples", samples)

        match self["ui.framework"]:
            case Backends.GLFW:
                self.glfw = self["ui.framework.class"].glfw
            case Backends.SDL:
                self.sdl3 = self["ui.framework.class"].sdl3
            case _:
                raise ValueError(f"Unknown UI Framework: {self['ui.framework']}")

        self.new("drawing.framework", drawing)
        self.new("drawing.framework.class", drawing_framework_map[drawing]())  # NOQA

        match self["drawing.framework"]:
            case Backends.SKIA:
                self.skia = importlib.import_module("skia")
            case _:
                raise ValueError(f"Unknown Drawing Framework: {self['drawing.framework']}")

        self.new("backend.framework", backend)

        match self["backend.framework"]:
            case Backends.OPENGL:
                self.opengl = importlib.import_module("OpenGL")
                self.opengl_GL = importlib.import_module("OpenGL.GL")
            case _:
                raise ValueError(f"Unknown Backend Framework: {self['backend.framework']}")

        self.windows = []
        self.is_alive: bool = False

        self._init_ui_framework()

    def _init_ui_framework(self):
        """According to attribute `ui.framework` to init ui framework"""
        match self.get("ui.framework"):
            case Backends.GLFW:
                self["ui.framework.class"].init(
                    error_callback=self.error, samples=self.get("ui.samples")
                )

    def update(self):
        """Update the Windows' UI and events"""

        input_mode: bool = True

        self.glfw.wait_events()
        windows = self.windows

        for window in windows:
            if window.is_visible and window.is_alive:
                window.update()
                if self.glfw.get_current_context():
                    self.glfw.swap_interval(1 if self.get("ui.is_vsync") else 0)  # 是否启用垂直同步

        if input_mode:
            self.glfw.poll_events()
        else:
            # if self._check_delay_events()
            self.glfw.wait_events()

    def mainloop(self):
        """Start mainloop.

        This method will start the manager event loop. It will continue running
        as long as the `is_alive` attribute is set to `True`.

        If no windows are added to the manager, a warning will be issued.
        """
        if not self.windows:
            warnings.warn(
                "At least one window is required to run application!",
            )

        self.is_alive: bool = True
        self["event.thread"].start()

        while self.is_alive:
            try:
                windows = self.windows
                if not windows:
                    self.quit()
                    break
                for window in windows:
                    if window.can_be_close():
                        window.destroy()
                self.update()
            except Exception as e:
                self.is_alive = False
                raise e

        self.cleanup()

    launch = run = mainloop

    def add_window(self, window):
        """Add a window to the manager.

        Args:
            window (charmy.widgets.WindowBase): The window to be added.
        """
        self.windows.append(window)

    def destroy_window(self, window):
        """Destroy a window from the manager.

        Args:
            window (charmy.widgets.WindowBase): The window to be destroyed.
        """
        windows = self.windows
        if window in windows:
            windows.remove(window)

    def cleanup(self) -> None:
        """Clean up resources."""
        match self.get("ui.framework"):
            case Backends.GLFW:
                for window in self.windows:
                    self.glfw.destroy_window(window.the_window)
                self.glfw.terminate()
        self.quit()

    def quit(self) -> None:
        """Quit the mainloop."""
        self.is_alive = False
        self["event.thread"].is_alive = False

    @staticmethod
    def error(error_code: typing.Any, description: bytes):
        """
        GLFW Error Callback

        :param error_code: GLFW error code
        :param description: Error description
        :return: None
        """
        raise f"GLFW Error {error_code}: {description.decode()}"


# Auto create Manager Object
uimap = {
    "GLFW": Backends.GLFW,
}
manager: CharmyManager = CharmyManager(
    id_=MANAGER_ID, ui=uimap[environ.get("UI_FRAMEWORK", "GLFW")]
)


def mainloop() -> None:
    """Start main loop."""
    try:
        manager.run()
    except Exception as e:
        raise e


def cquit():  # NOQA
    """Quit the main loop"""
    manager.quit()



# Copyright (C) 2026 XiangQinxi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#    http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: XiangQinxi
# Contributors: XiangQinxi, rgzz666, littlewhitecloud

"""A modern GUI library."""

from .const import *

# from .drawing import *
from .event import Event, EventHandling, EventTask, WorkingThread
from .object import CharmyObject
from .pos import Pos
from .rect import Rect
from .size import Size
from .styles import *
from .widgets import *



import importlib.util
import sys
import typing

from .const import Backends
from .event import Event
from .pos import Pos
from .size import Size

from abc import ABC, abstractmethod


# region Window
class WindowFramework(ABC):
    """The base class of WindowFramework."""

    @abstractmethod
    def init(self, **kwargs) -> None:
        """Init the window framework

        Args:
            **kwargs: Keyword arguments.

        Returns:
            None
        """
        ...

    @abstractmethod
    def create(self, size, title, **kwargs) -> dict[str, typing.Any]:
        """Create a window

        Args:
            size: Size of the window
            title: Title of the window
            **kwargs: Keyword arguments. GLFW:
                samples: Number of samples for multisampling. Default is 4.
                error_callback: Error callback function. Default is None.

        Returns:
            dict[str, typing.Any]: A dictionary containing the created window including 'window' and 'pos'
        """
        ...

    @abstractmethod
    def create_event_bounds(self, the_window, window_class, **kwargs):
        """Create event bounds for the window

        Args:
            the_window: The window to create event bounds for
            window_class: The window class to create event bounds for
            **kwargs: Keyword arguments.

        Returns:
            None
        """
        ...

    @abstractmethod
    def destroy(self, the_window) -> None:
        """Destroy the window

        Args:
            the_window: The window to destroy with

        Returns:
            None
        """
        ...

    @abstractmethod
    def set_size(self, the_window, size: Size) -> None:
        """Set the size of the window

        Args:
            the_window: The window to set size for
            size: Size to set

        Returns:
            None
        """
        ...

    @abstractmethod
    def set_pos(self, the_window, pos: Pos | tuple[int, int]) -> None:
        """Set the position of the window

        Args:
            the_window: The window to set position for
            pos: Position to set

        Returns:
            None
        """
        ...

    @abstractmethod
    def set_title(self, the_window, title: str) -> None:
        """Set the title of the window

        Args:
            the_window: The window to set title for
            title: Title to set (Default value = 'Charmy GUI')

        Returns:
            None
        """
        ...

    @abstractmethod
    def get_mouse_pos(self, the_window) -> tuple[int, int]:
        """Get the mouse position of the window

        Args:
            the_window: The window to get mouse position for

        Returns:
            tuple[int, int]: The mouse position of the window
        """
        ...

    def make_context_current(self, the_window, **kwargs) -> dict[str, typing.Any]:
        """Make the context current for the window

        Args:
            the_window: The window to make context current for

        Returns:
            None
        """
        ...

    @abstractmethod
    def swap_buffers(self, the_window) -> None:
        """Swap the buffers of the window

        Args:
            the_window: The window to swap buffers for

        Returns:
            None
        """
        ...


window_framework_map: dict[Backends, WindowFramework] = {}


class GLFW(WindowFramework):
    def __init__(self):
        self.glfw = importlib.import_module("glfw")

    def init(self, **kwargs) -> None:
        if not self.glfw.init():
            raise self.glfw.GLFWError("Init failed")

        self.glfw.window_hint(self.glfw.STENCIL_BITS, 8)
        self.glfw.window_hint(self.glfw.TRANSPARENT_FRAMEBUFFER, True)
        self.glfw.window_hint(self.glfw.WIN32_KEYBOARD_MENU, True)
        self.glfw.window_hint(self.glfw.COCOA_RETINA_FRAMEBUFFER, True)
        self.glfw.window_hint(self.glfw.SAMPLES, kwargs.get("samples", 4))
        if kwargs.get("error_callback", None):
            self.glfw.set_error_callback(kwargs["error_callback"])
        ...

    def create(self, size, title, **kwargs) -> dict[str, typing.Any]:
        self.glfw.window_hint(
            self.glfw.CONTEXT_RELEASE_BEHAVIOR, self.glfw.RELEASE_BEHAVIOR_NONE
        )  # mystery optimize
        self.glfw.window_hint(self.glfw.STENCIL_BITS, 8)
        self.glfw.window_hint(self.glfw.COCOA_RETINA_FRAMEBUFFER, self.glfw.TRUE)  # macOS
        self.glfw.window_hint(self.glfw.SCALE_TO_MONITOR, self.glfw.TRUE)  # Windows/Linux

        # see https://www.glfw.org/faq#macos
        if sys.platform.startswith("darwin"):
            self.glfw.window_hint(self.glfw.CONTEXT_VERSION_MAJOR, 3)
            self.glfw.window_hint(self.glfw.CONTEXT_VERSION_MINOR, 2)
            self.glfw.window_hint(self.glfw.OPENGL_FORWARD_COMPAT, True)
            self.glfw.window_hint(self.glfw.OPENGL_PROFILE, self.glfw.OPENGL_CORE_PROFILE)
        else:
            if kwargs.get("fha", True):
                self.glfw.window_hint(self.glfw.OPENGL_FORWARD_COMPAT, True)
                self.glfw.window_hint(self.glfw.CLIENT_API, self.glfw.OPENGL_API)
                self.glfw.window_hint(self.glfw.CONTEXT_VERSION_MAJOR, 3)
                self.glfw.window_hint(self.glfw.CONTEXT_VERSION_MINOR, 3)
                self.glfw.window_hint(self.glfw.OPENGL_PROFILE, self.glfw.OPENGL_CORE_PROFILE)

        window = self.glfw.create_window(size["width"], size["height"], title, None, None)
        if not window:
            raise RuntimeError("Can't create window")

        return {"pos": self.glfw.get_window_pos(window), "window": window}

    def create_event_bounds(self, the_window, window_class, **kwargs) -> None:
        self.glfw.set_window_size_callback(
            the_window,
            lambda w, width, height: window_class.trigger(
                Event(window_class, "resize", width=width, height=height)
            ),
        )
        self.glfw.set_window_pos_callback(
            the_window,
            lambda w, root_x, root_y: window_class.trigger(
                Event(window_class, "move", x_root=root_x, y_root=root_y)
            ),
        )

        def _enter(w, entered: int):
            if entered:
                event_type = "mouse_enter"
            else:
                event_type = "mouse_leave"
            window_class.trigger(Event(window_class, event_type, the_window=w))

        self.glfw.set_cursor_enter_callback(
            the_window,
            _enter,
        )
        self.glfw.set_cursor_pos_callback(
            the_window,
            lambda w, x, y: window_class.trigger(
                Event(window_class, "mouse_move", x=x, canvas_x=x, y=y, canvas_y=y)
            ),
        )

        def _mouse(w, button, action, mods):  # NOQA
            if action == self.glfw.PRESS:
                event_type = "mouse_press"
            elif action == self.glfw.RELEASE:
                event_type = "mouse_release"
            else:
                return
            button_map = {
                self.glfw.MOUSE_BUTTON_LEFT: "left",
                self.glfw.MOUSE_BUTTON_RIGHT: "right",
                self.glfw.MOUSE_BUTTON_MIDDLE: "middle",
            }
            mods_map = {
                self.glfw.MOD_SHIFT: "shift",
                self.glfw.MOD_CONTROL: "control",
                self.glfw.MOD_ALT: "alt",
                self.glfw.MOD_SUPER: "super",
                self.glfw.MOD_NUM_LOCK: "num_lock",
                self.glfw.MOD_CAPS_LOCK: "caps_lock",
            }
            window_class.trigger(
                Event(
                    window_class,
                    event_type,
                    button=button_map.get(button, None),
                    mods=mods_map.get(mods, None),
                )
            )

        self.glfw.set_mouse_button_callback(the_window, _mouse)

    def destroy(self, the_window) -> None:
        self.glfw.destroy_window(the_window)

    def set_size(self, the_window, size: Size | tuple[int, int]) -> None:
        if isinstance(size, tuple):
            self.glfw.set_window_size(the_window, size[0], size[1])
        else:
            self.glfw.set_window_size(the_window, size["width"], size["height"])

    def set_pos(self, the_window, pos: Pos | tuple[int, int]) -> None:
        if isinstance(pos, tuple):
            self.glfw.set_window_pos(the_window, pos[0], pos[1])
        else:
            self.glfw.set_window_pos(the_window, pos["x"], pos["y"])

    def set_title(self, the_window, title: str) -> None:
        self.glfw.set_title(the_window, title)

    def get_mouse_pos(self, the_window) -> tuple[int, int]:
        return self.glfw.get_cursor_pos(the_window)

    def make_context_current(self, the_window, **kwargs) -> None:
        self.glfw.make_context_current(the_window)

    def swap_buffers(self, the_window) -> None:
        self.glfw.swap_buffers(the_window)

    def can_be_closed(self, the_window) -> bool:
        return self.glfw.window_should_close(the_window)


if importlib.util.find_spec("glfw") is not None:
    window_framework_map[Backends.GLFW] = GLFW  # NOQA
# endregion

from .rect import Rect


# region Drawing
class DrawingFramework:
    """The base class of DrawingFramework."""

    def draw_rect(self, canvas, rect: Rect, radius: int = 0, bg=None, bd=None):
        """Draw a rectangle

        Args:
            canvas (Canvas): The canvas to draw
            rect (Rect): The rect area to draw
            radius (int, optional): The radius of the rectangle. Defaults to 0.
            bg (charmy.styles.color.Color, optional): The background color of the rectangle. Defaults to None.
            bd (charmy.styles.color.Color, optional): The border color of the rectangle. Defaults to None.
        """
        ...


drawing_framework_map: dict[Backends, DrawingFramework] = {}


class SKIA:
    def __init__(self):
        self.skia = importlib.import_module("skia")

    def draw_rect(self, canvas, rect: Rect, radius: int | float = 8, bg=None, bd=None):  # noqa
        # from skia import RRect, Rect, Canvas, Paint
        # Paint()
        # canvas: Canvas

        if bg is None:
            bg = {}
        canvas.drawRoundRect(
            rect=self.skia.Rect.MakeXYWH(rect["x"], rect["y"], rect["width"], rect["height"]),
            rx=radius,
            ry=radius,
            paint=self.skia.Paint(
                Color=bg.get("color_object", self.skia.ColorBLUE),
                Style=self.skia.Paint.kFill_Style,
            ),
        )


if importlib.util.find_spec("skia") is not None:
    drawing_framework_map[Backends.SKIA] = SKIA  # NOQA

# endregion



"""Charmy constants."""

from enum import Enum


class ID(Enum):
    """ID is an enum to store object ID.

    AUTO: Auto generate ID.
    NONE: No ID.
    """

    AUTO = 0
    NONE = 1


class Backends(Enum):
    """UIFrame is an enum to store UI frame.

    GLFW, SDL, OPENGL, SKIA are consts representing those backends
    """

    GLFW = glfw = "GLFW"
    SDL = sdl = "SDL"
    OPENGL = opengl = "OPENGL"
    SKIA = skia = "SKIA"


class DrawingMode(Enum):
    """DrawingMode is an enum to store drawing mode.
    IMMEDIATE(immediate): IMMEDIATE is an enum to store drawing mode.
    RETAINED(retained): RETAINED is an enum to store drawing mode.
    """

    IMMEDIATE = immediate = "IMMEDIATE"
    RETAINED = retained = "READIED"


MANAGER_ID = "main_manager"



from ..const import Backends
from .container import Container
from .windowbase import WindowBase


class Window(WindowBase, Container):
    """Window class."""

    def __init__(self, *args, **kwargs):
        WindowBase.__init__(self, *args, **kwargs)
        self.new("children", [])
        match self["drawing.framework"]:
            case Backends.SKIA:
                self.set(
                    "ui.draw_func", self.skia_draw_func
                )  # When `draw` is called, trigger `ui.draw_func`

    def skia_draw_func(self, canvas):
        """Draw function for Skia."""
        canvas.clear(self.skia.ColorGRAY)
        self.draw_children(canvas)



import typing

from ..const import MANAGER_ID, Backends, DrawingMode
from ..event import Event, EventHandling
from ..object import CharmyObject
from ..pos import Pos
from ..size import Size
from .cmm import CharmyManager


class WindowBase(EventHandling, CharmyObject):
    """WindowBase is a base class for window.

    Args:
        parent: The CharmyManager object,
        title: The title of the window,
        size: The size of the window,
        fha: Whether to force hardware acceleration
        drawing_mode: The drawing mode of the window,
    """

    def __init__(
        self,
        parent: CharmyManager | None = None,
        *,
        title: str = "Charmy GUI",
        size: tuple[int, int] = (100, 100),
        fha: bool = True,
        drawing_mode: DrawingMode = DrawingMode.RETAINED,
    ):
        super().__init__()

        # Init parent attribute
        if parent is None:
            parent = self.get_obj(MANAGER_ID)

        if isinstance(parent, CharmyManager):
            self.manager = parent
        elif isinstance(parent, WindowBase):
            self.manager = parent.manager
        self.manager.add_window(self)

        # Init Attributes
        self.new(
            "ui.framework", self._get_ui_framework(), get_func=self._get_ui_framework
        )  # The UI Framework
        self.new(
            "ui.framework.class",
            self._get_ui_framework_class(),
            get_func=self._get_ui_framework_class,
        )
        self.new(
            "ui.is_vsync", self._get_ui_is_vsync(), get_func=self._get_ui_is_vsync
        )  # Whether to enable VSync

        # """
        match self["ui.framework"]:
            case Backends.GLFW:
                self.glfw = self.manager.glfw
            case Backends.SDL:
                self.sdl3 = self.manager.sdl3
            case _:
                raise ValueError(f"Unknown UI Framework: {self['ui.framework']}")
        # """

        self.new("ui.draw_func", None)

        self.new(
            "drawing.framework",
            self._get_drawing_framework(),
            get_func=self._get_drawing_framework,
        )
        self.new("drawing.mode", drawing_mode)

        match self["drawing.framework"]:
            case Backends.SKIA:
                self.skia = self.manager.skia
                self.new("drawing.surface", None)
            case _:
                raise ValueError(f"Unknown Drawing Framework: {self['drawing.framework']}")

        self.new(
            "backend.framework",
            self._get_backend_framework(),
            get_func=self._get_backend_framework,
        )

        match self["backend.framework"]:
            case Backends.OPENGL:
                self.opengl = self.manager.opengl
                self.opengl_GL = self.manager.opengl_GL
                self.new("backend.context", None)
            case _:
                raise ValueError(f"Unknown Backend Framework: {self['backend.framework']}")

        self.new("pos", Pos(0, 0))  # Always (0, 0)
        self.new("canvas_pos", Pos(0, 0))  # Always (0, 0)
        self.new("root_pos", Pos(0, 0), set_func=self._set_pos)  # The position of the window
        self.new("size", Size(size[0], size[1]), set_func=self._set_size)  # The size of the window
        self.new("title", title, set_func=self._set_title)  # The title of the window

        self.is_dirty: bool = True
        self.is_force_hardware_acceleration: bool = fha
        self.is_visible: bool = False  # Is the window visible
        self.is_alive: bool = False  # Is the window alive

        self.the_window = self.create()  # GLFW/SDL Window

        self.create_event_bounds()

        self.bind("move", self._on_move)
        self.bind("resize", self._on_resize)

    def create(self):
        """Create the window."""
        arg = self["ui.framework.class"].create(
            size=self.get("size", skip=True),
            title=self.get("title"),
            fha=self.is_force_hardware_acceleration,
        )

        _root_point = self.get("root_pos", skip=True)(arg["pos"][0], arg["pos"][1])

        self.is_visible = True
        self.is_alive = True

        return arg["window"]

    def create_event_bounds(self):
        """Create event bounds."""
        self["ui.framework.class"].create_event_bounds(
            the_window=self.the_window, window_class=self
        )

    def update(self):
        """Update the window. When is_dirty is True, draw the window."""
        if self.is_visible:
            match self["drawing.mode"]:
                case DrawingMode.IMMEDIATE:
                    self.draw()
                case DrawingMode.RETAINED:
                    if self.is_dirty:
                        self.draw()
                        self.is_dirty = False

    import contextlib

    @contextlib.contextmanager
    def skia_surface(self, arg: typing.Any):
        """Create a Skia surface for the window.

        :param arg: GLFW or SDL2 Window/Surface
        :return: Skia Surface
        """
        match self["ui.framework"]:
            case Backends.GLFW:
                if not self.glfw.get_current_context() or self.glfw.window_should_close(arg):
                    yield None
                    return

                match self["backend.framework"]:
                    case Backends.OPENGL:
                        match self["drawing.framework"]:
                            case Backends.SKIA:
                                self["backend.context"] = self.skia.GrDirectContext.MakeGL()
                                fb_width, fb_height = self.glfw.get_framebuffer_size(arg)
                                backend_render_target = self.skia.GrBackendRenderTarget(
                                    fb_width,
                                    fb_height,
                                    0,
                                    0,
                                    self.skia.GrGLFramebufferInfo(0, self.opengl_GL.GL_RGBA8),
                                )
                                surface = self.skia.Surface.MakeFromBackendRenderTarget(
                                    self["backend.context"],
                                    backend_render_target,
                                    self.skia.kBottomLeft_GrSurfaceOrigin,
                                    self.skia.kRGBA_8888_ColorType,
                                    self.skia.ColorSpace.MakeSRGB(),
                                )
                                self["backend.context"].setResourceCacheLimit(16 * 1024 * 1024)

                                if surface is None:
                                    raise RuntimeError("Failed to create Skia surface")

                                yield surface

            case Backends.SDL:
                import ctypes

                width, height = arg.w, arg.h
                pixels_ptr = arg.pixels
                pitch = arg.pitch

                # SDL 像素包装成 buffer
                buf_type = ctypes.c_uint8 * (pitch * height)
                buf = buf_type.from_address(pixels_ptr)  # NOQA

                imageinfo = self.skia.ImageInfo.MakeN32Premul(width, height)  # NOQA
                surface = self.skia.Surface.MakeRasterDirect(imageinfo, buf, pitch)

                if surface is None:
                    raise RuntimeError("Failed to create Skia surface")

                yield surface  # ⚠️ 必须用 yield，不要 return

    def draw(self, event: Event | None = None) -> None:  # NOQA
        """Draw the window.

        Args:
            event (CEvent, optional): The event that triggered the draw. Defaults to None.
        """
        if self.is_visible:
            # Set the current context for each arg
            # 【为该窗口设置当前上下文】
            match self["ui.framework"]:
                case Backends.GLFW:
                    self["ui.framework.class"].make_context_current(self.the_window)

                    match self["drawing.framework"]:
                        case Backends.SKIA:
                            # Create a Surface and hand it over to this arg.
                            # 【创建Surface，交给该窗口】
                            with self.skia_surface(self.the_window) as self[  # NOQA
                                "drawing.surface"
                            ]:  # NOQA
                                if self["drawing.surface"]:
                                    with self["drawing.surface"] as canvas:
                                        # Determine and call the drawing function of this arg.
                                        # 【判断并调用该窗口的绘制函数】
                                        if self["ui.draw_func"]:
                                            self["ui.draw_func"](canvas)

                                    self["drawing.surface"].flushAndSubmit()
                case "sdl2":
                    import sdl2  # NOQA

                    surface = sdl2.SDL_GetWindowSurface(self.the_window).contents

                    with self.skia_surface(surface) as self["drawing.surface"]:  # NOQA
                        if self["drawing.surface"]:
                            with self["drawing.surface"] as canvas:
                                if self["ui.draw_func"]:
                                    self["ui.draw_func"](canvas)

                    sdl2.SDL_UpdateWindowSurface(self.the_window)  # NOQA

            if self.is_alive:
                self["ui.framework.class"].swap_buffers(self.the_window)

        if self["backend.context"]:
            self["backend.context"].freeGpuResources()
            self["backend.context"].releaseResourcesAndAbandonContext()
        # for child in self.children:
        #    child.need_redraw = False
        self.trigger(Event(self, "draw"))

    def dirty(self):
        """Set the dirty flag."""
        self.is_dirty = True

    def cancel_dirty(self):
        """Cancel the dirty flag."""
        self.is_dirty = False

    def destroy(self) -> None:
        """Destroy the window.

        :return: None
        """
        # self._event_init = False
        # print(self.id)
        self.manager.destroy_window(self)
        try:
            self["ui.framework.class"].destroy(the_window=self.the_window)
        except TypeError:
            pass
        finally:
            self.is_alive = False
            # self.draw_func = None
            self.the_window = None  # Clear the reference

    def can_be_close(self, value: bool | None = None) -> typing.Self | bool:
        """Set whether the window can be closed.

        Prevent users from closing the window, which can be used in conjunction with prompts like "Save before closing?"

        >>> def delete(_: SkEvent):  # NOQA
        >>>     window.can_be_close(False)  # NOQA
        >>> window.bind("delete_window", delete)  # NOQA


        :param value: Whether the window can be closed
        :return: None
        """
        match self.get("ui.framework"):
            case Backends.GLFW:
                import glfw

                if value is not None:
                    glfw.set_window_should_close(self.the_window, value)
                    return self
                else:
                    if self.the_window:
                        return glfw.window_should_close(self.the_window)
                    else:
                        return False
        return True

    # region Getters and Setters
    def _get_ui_framework(self):
        return self.manager.get("ui.framework")

    def _get_ui_framework_class(self):
        return self.manager.get("ui.framework.class")

    def _get_drawing_framework(self):
        return self.manager.get("drawing.framework")

    def _get_backend_framework(self):
        return self.manager.get("backend.framework")

    def _get_ui_is_vsync(self):
        return self.manager.get("ui.is_vsync")

    def _set_size(self, size: Size | tuple[int, int]) -> None:
        """Set the size of the window.

        Args:
            size (Size | tuple[int, int]): Size to set.
        Returns:
            None
        """
        self["ui.framework.class"].set_size(the_window=self.the_window, size=size)

    def resize(self, size: Size | tuple[int, int]) -> None:
        """Resize the window to the given size.

        Args:
            size: Size to resize.
        Returns:
            None
        """
        self.set("size", size)

    def _set_pos(self, pos: Pos | tuple[int, int]) -> None:
        """Set the position of the window.

        Args:
            pos (CPOS | tuple[int, int]): Position to set
        Returns:
            None
        """
        self["ui.framework.class"].set_pos(the_window=self.the_window, pos=pos)

    def move(self, pos: Pos | tuple[int, int]) -> None:
        """Move the window to the given position.

        Args:
            pos: Position to move
        Returns:
            None
        """
        self.set("root_pos", pos)

    def _set_title(self, title: str) -> None:
        """Set the title of the window.

        Args:
            title (str): Title to set
        Returns:
            None
        """
        self["ui.framework.class"].set_title(the_window=self.the_window, title=title)

    # endregion

    # region Events

    def _on_move(self, event: Event):
        """Handle the move event.

        Args:
            event (Event): The move event.
        """
        _root_point = self.get("root_pos", skip=True)
        _root_point.set("x", event["x_root"])
        _root_point.set("y", event["y_root"])

    def _on_resize(self, event: Event):
        """Handle the resize event.

        Args:
            event (Event): The resize event.
        """
        _size = self.get("size", skip=True)
        _size(event["width"], event["height"])
        self.dirty()

    # endregion



from __future__ import annotations as _

import inspect
import json
import os
import pathlib
import re
import typing
import warnings

from . import color

if typing.TYPE_CHECKING:
    # from .。widgets.widget import CWidget
    pass


class CStyleNotFoundError(NameError):
    """Will be raised when a theme is not found."""

    pass


class CTheme:
    """Theme class for CWindow and CWidgets.

    Example
    -------
    .. code-block:: python
        my_theme = CTheme({<Some styles>})
        my_sub_theme = CTheme(parent="default.light")
        my_external_theme = CTheme().load_from_file("./path/to/a/theme.json")
    This shows examples of creating themes, either from a json, a parent theme or a file.

    .. code-block:: python
        all_themes = CTheme.loaded_themes
        internal_themes = CTheme.INTERNAL_THEMES
        default_theme = CTheme.DEFAULT_THEME
    This shows getting all loaded themes, internal themes, and the default theme.

    .. code-block:: python
        default_light_theme = CTheme.find_loaded_theme("default.light")
        if CTheme.validate_theme_existed("default.light"):
            print("Default light theme exists!")
    This shows finding a theme and checking if it exists

    """

    loaded_themes: list["CTheme"] = []
    INTERNAL_THEME_DIR = pathlib.Path(__file__).parent.parent / "resources" / "themes"
    INTERNAL_THEMES: dict[str, "CTheme"] = {}
    DEFAULT_THEME: "CTheme"
    DEFAULT_THEME_FILENAME: str = "light"

    @classmethod
    def _load_internal_themes(cls):
        """Load internal themes. Should be run once at import, see the end of this file."""
        # Load default (ROOT) theme
        CTheme.DEFAULT_THEME = CTheme({}).load_from_file(
            CTheme.INTERNAL_THEME_DIR / f"{CTheme.DEFAULT_THEME_FILENAME}.json"
        )

        # Load other internal themes
        for file in os.listdir(CTheme.INTERNAL_THEME_DIR):
            if file == f"{CTheme.DEFAULT_THEME_FILENAME}.json":
                # For default theme, no need to reload it
                CTheme.INTERNAL_THEMES[CTheme.DEFAULT_THEME.name] = CTheme.DEFAULT_THEME
                continue
            _ = CTheme({}).load_from_file(CTheme.INTERNAL_THEME_DIR / file)
            CTheme.INTERNAL_THEMES[_.name] = _

    @classmethod
    def find_loaded_theme(cls, theme_name: str) -> "CTheme | typing.Literal[False]":
        """Search for a loaded theme by name, returns the CTheme object if found, or False if not.

        Example
        -------
        .. code-block:: python
            default_theme = CTheme.find_loaded_theme("default.light")
        This returns the CTheme object of the default theme to `default_theme`.

        :param theme_name: Name of the theme to load
        :return: The CTheme object if found, otherwise False
        """
        for theme in cls.loaded_themes:
            if theme.name == theme_name:
                return theme
        return False

    @classmethod
    def validate_theme_existed(cls, theme_name: str) -> bool:
        """Validate if the theme with given name existed and loaded.

        Example
        -------
        .. code-block:: python
            CTheme.validate_theme_existed("default.light")
        This returns if the theme `default.light` is loaded.

        :param theme_name: Name of the theme to validate
        :return: If the theme loaded
        """
        return CTheme.find_loaded_theme(theme_name) != False  # ☝🤓

    def __init__(
        self, styles: dict | None = None, parent: typing.Union["CTheme", None] = None
    ) -> None:
        """Theme for CWindow and CWidgets.

        Example
        -------
        .. code-block:: python
            my_theme = CTheme({<Some styles>})
            my_sub_theme = CTheme(parent="default.light")
            my_external_theme = CTheme().load_from_file("./path/to/a/theme.json")
        This shows examples of creating themes, either from a json, a parent theme or a file.

        :param styles: Styles of the theme
        :param parent: Parent theme
        """

        self.name: str = f"untitled.{len(CTheme.loaded_themes) + 1}"
        self.friendly_name = f"Untitled theme {len(CTheme.loaded_themes) + 1}"
        # friendly_name感觉有点多余? ——Little White Cloud
        # Keep it 4 now currently. ——rgzz666
        self.parent: typing.Union["CTheme", None] = parent
        self.children = []
        self.is_special = False

        if styles is None:
            self.styles: dict = CTheme.DEFAULT_THEME.styles
        else:
            self.styles: dict = styles
        self.color_palette = {}

        CTheme.loaded_themes.append(self)
        return

    def load_from_file(self, file_path: str | pathlib.Path) -> "CTheme":
        """Load styles to theme from a file.

        Example
        -------
        .. code-block:: python
            my_theme = CTheme().load_from_file("./path/to/a/theme.json")
            my_theme.load_from_file("./path/to/another/theme.json")

        This shows loading a theme to `my_theme` from the theme file at `./path/to/a/theme.json`,
        and change it to theme from `./path/to/another/theme.json` later.

        :param file_path: Path to the theme file
        :return self: The CTheme itself
        """
        # Change path string into pathlib Path
        if type(file_path) is str:
            file_path = pathlib.Path(file_path)
        # Get path where lies codes calling this function (to support relative path)
        if not file_path.is_absolute():
            frame = inspect.currentframe()
            outer_frame = inspect.getouterframes(frame)[1]
            caller_file = pathlib.Path(outer_frame.filename).parent
            file_path = (caller_file / file_path).resolve()
        # We need a file to load from file \o/ \o/ \o/
        with open(file_path, mode="r", encoding="utf-8") as f:
            style_raw = f.read()
            theme_data = json.loads(style_raw)
            if (search_result := CTheme.find_loaded_theme(theme_data["name"])) != False:
                # If name already occupied, meaning the theme might already be loaded
                # (or just simply has an occupied name)
                warnings.warn(
                    f"Theme <{theme_data['name']}> already loaded or existed.",
                    RuntimeWarning,
                )
                return search_result

        return self.load_from_json(theme_data)

    def load_from_json(self, theme_data: dict) -> "CTheme":
        """Load all data (including metadata) to the theme.

        Example
        -------
        .. code-block:: python
            my_theme = CTheme().load_from_json({<Some JSON theme data>})
            my_theme.load_from_json({<Some JSON theme data>})
        This shows loading a theme to `my_theme` from json data, and change it to theme from
        another json later.

        :param theme_data: dict that contains the theme data
        :return self: The CTheme itself
        """
        # Type check
        EXPECTED_DATA_TYPE = {
            "styles": dict,
            "color_palette": dict,
            "name": str,
            "friendly_name": str,
            "base": str,
        }
        for item in EXPECTED_DATA_TYPE.keys():
            if type(theme_data[item]) != EXPECTED_DATA_TYPE[item]:
                theme_name = (
                    theme_data["name"] if type(theme_data["name"]) is str else "(Type error)"
                )
                warnings.warn(
                    f"Error data type of <{item}> in theme data that is about to be loaded. "
                    f"Expected {EXPECTED_DATA_TYPE[item]} but got {type(item)}. The json data with "
                    f"theme named <{theme_name}> will not be loaded to the theme <{self.name}>",
                    ResourceWarning,
                )
                return self
        # Load data
        self.styles = theme_data["styles"].copy()
        self.color_palette = theme_data["color_palette"].copy()
        # Load Metadata
        self.rename(theme_data["name"], theme_data["friendly_name"])
        self.set_parent(theme_data["base"])

        return self

    def load_styles_from_json(self, style_json: dict) -> "CTheme":
        """Load styles to theme from a given dict.

        Example
        -------
        .. code-block:: python
            my_theme = CTheme().load_styles_from_json({<Some styles>})
            my_theme.load_from_json({<Some styles>})
        This shows loading styles data to `my_theme` from json data, and change its styles from
        that stored in another json later.

        :param style_json: dict that contains the styles
        :return self: The CTheme itself
        """
        # This just fucks everything from source json into the styles
        self.styles = style_json.copy()
        return self

    def set_parent(self, new_parent: str) -> "CTheme":
        """Set the parent for the theme via string stored in theme json.

        Example
        -------
        .. code-block:: python
            my_theme = CTheme().set_parent("DEFAULT")
            my_theme.set_parent("default.dark")
        The first line shows setting the parent of `my_theme` to the default theme, which is
        suggested for third-party themes that act as a root. The second line shows setting the
        parent of `my_theme` to `default.dark` after its creation.

        Parent Name
        -----------
        - `ROOT` means the theme does not have a parent. This is not recommended for third-party
          themes as fallback mechanisms will all stop working, use `DEFAULT` instead.
        - `DEFAULT` means the parent of the theme is the default internal theme.

        If the parent name is none of above, it should be the theme of the name and will be set as
        parent directly. However, if the theme specified is not yet loaded, parent will fall back
        to `DEFAULT`.

        :param parent_name: Name of the new parent, or the CTheme object of it
        :return self: The CTheme itself
        """
        match new_parent:
            case "ROOT":
                # If root theme, then no parent (which means no fallback)
                self.parent = None
            case "DEFAULT":
                # If default theme for parent, then... hmm... default theme for parent... 🤔
                self.parent = self.DEFAULT_THEME
            case CTheme():
                # If is a CTheme object, set it as parent
                self.parent = new_parent
            case _:
                # If else, find the theme
                search_result = CTheme.find_loaded_theme(new_parent)
                if search_result != False:
                    self.parent = search_result
                else:
                    # When not found, fallback to default for parent
                    warnings.warn(
                        f"Parent theme specified with name <{new_parent}> is not yet loaded. "
                        "Will fall back to <DEFAULT> for parent instead."
                    )
                    self.set_parent("DEFAULT")
        if isinstance(self.parent, CTheme):
            self.parent.children.append(self)
        return self

    def rename(self, new_name: str, friendly_name: str) -> "CTheme":
        """Rename the theme.

        Example
        -------
        .. code-block:: python
            my_theme = Theme().rename("theme.name")
            my_theme.rename("i_hate.that_name")
        This shows renaming `my_theme` to `theme.name`, and renaming it to `i_hate.that_name` after
        its creation.

        :param friendly_name:
        :param new_name: The new name for the theme
        :return self: The CTheme itself
        """
        if not CTheme.validate_theme_existed(new_name):
            # If name not occupied, then rename self.
            self.name = new_name
            self.friendly_name = friendly_name  # 🤔
        else:
            # Otherwise stop this.
            warnings.warn(
                f"Theme name <{new_name}> occupied. Rename for <{self.name}> is canceled."
            )
        return self

    @staticmethod
    def parse_selector(selector: str) -> list[str]:
        """Parse styles selector.

        This is a selector parser mainly used by internal functions.

        Example
        -------
        See -> :func:`get_style` in source code.

        Selector
        --------
        - `<Widget>` indicates the styles of Widget at rest state, e.g. `CButton`.
        - `<Widget>:<state>` indicates the styles of the state of Widget, e.g. `CButton:hover`.
        - `<Widget>:ITSELF` indicates the styles of the widget, e.g. `CButton.ITSELF`.
          Note that this is not available everywhere.

        :param selector: The selector string
        :return: Parsed selector, levels in a list
        """
        # Validate if selector valid
        if not re.match("[a-zA-Z0-9-_.:,]+", selector):
            raise ValueError(f"Invalid styles selector [{selector}].")
        # Handling
        result: list[str] = []
        if ":" in selector:
            colon_parsed = selector.split(":")
            if len(colon_parsed) > 2:  # Validation
                raise ValueError(f"Invalid styles selector [{selector}].")
            # # Deprecated code of multi-state
            # if "," in colon_parsed[1]:
            #     # If has more than one state specified:
            #     result[1] = colon_parsed[1].split(",")
            # else:
            #     # Otherwise, we still make it a list type
            #     result[1] = [colon_parsed[1]]
            if colon_parsed[1] == "ITSELF":
                # For ITSELF selectors
                result = [result[0]]
            result = colon_parsed
        else:
            result = [selector, "rest"]
        # Return the parsed selector
        return result

    @typing.overload
    def select(
        self,
        selector: str | list,
        *,
        copy: bool = ...,
        fallback: bool = ...,
        make_path: bool = ...,
        allow_not_found: typing.Literal[True] = True,
    ) -> dict | None: ...

    @typing.overload
    def select(
        self,
        selector: str | list,
        *,
        copy: bool = ...,
        fallback: bool = ...,
        make_path: bool = ...,
        allow_not_found: typing.Literal[False] = False,
    ) -> dict: ...

    def select(
        self,
        selector: str | list,
        *,
        copy: bool = True,
        fallback: bool = True,
        make_path: bool = False,
        allow_not_found: bool = True,
    ) -> dict | None:
        """Get styles config using a selector.

        Example
        -------
        .. code-block:: python
            my_theme = CTheme()
            my_style = my_theme.get_style("CButton:hover")
            my_theme.get_style("CButton:hover", copy=False)["background"] = (255, 0, 0, 255)
        This shows getting style json of CButton at hover state and setting its background to red.

        :param selector: The selector string, indicating which styles to get
        :param copy: Whether to copy a new styles json, otherwise returns the styles itself
        :param fallback: Whether to enable fallback machanism, defaults to True
        :param make_path: Whether to create the path if not found instead of throwing errors,
                          will force disable fallback when enabled, defaults to False
        :param allow_not_found: True to return None when style not found, otherwise raise error in
                                such cases
        :return result: The style dict
        """
        # Params handling related stuff
        if make_path:
            # Force disable fallback when make_path enabled
            fallback = False
        # First, set the result to all styles
        result = self.styles
        if not selector:
            # If no selector is provided, then return all styles, so do nothing
            pass
        else:
            try:
                # To get a parsed selector
                selector_parsed = (
                    self.parse_selector(selector) if type(selector) is str else selector
                )  # If is already parsed list
                # Validate if selector exists in theme
                _ = self.styles
                for selector_level in selector_parsed:
                    # # Deprecated code of multi-state
                    # if type(selector_level) is list:
                    #     # If is more than one state specified
                    #     selector_level = selector_level[0] # Then take the first during checking
                    if selector_level not in _:
                        if isinstance(self.parent, CTheme) and fallback:
                            # If parent exists, then fallback
                            return self.parent.select(
                                selector,
                                copy=copy,
                                fallback=fallback,
                                make_path=make_path,
                                allow_not_found=allow_not_found,
                            )
                        else:
                            # If is root theme, then go fuck ur selector
                            if make_path:
                                # If make_path enabled, make the path
                                _[selector_level] = {}
                            else:
                                # Otherwise u should really go fuck ur selector
                                if allow_not_found:
                                    return None
                                else:
                                    raise CStyleNotFoundError(
                                        "Cannot find styles with selector " f"[{selector}]"
                                    )
                    _ = _[selector_level]  # Heading to the next level
            except CStyleNotFoundError:
                # If this fails, then the selector is invalid
                raise CStyleNotFoundError(
                    f"Style [{selector}] is not exsited in the default theme. Check your selector!"
                )

            for selector_level in selector_parsed:
                # e.g. result = styles["CButton"]
                # result = styles["CButton"]["hover"]
                result = result[selector_level]

        if copy and type(result) is dict:
            return result.copy()
        else:
            return result

    def get_style_attr(
        self, selector: str | list[str], attr_name: str, allow_not_found: bool = True
    ) -> typing.Any:
        """Get style attribute _value.

        Example
        -------
        .. code-block:: python
            my_theme = CTheme()
            button_background = my_theme.get_style_attr("CButton:rest", "background")
        This shows getting the _value of `background` attribute from styles of `CButton` at `rest`
        state.

        Fallback Mechanism
        ------------------
        - The program first tries to get attribute from the style indicated by the selector.
        - If fails, which means that the attribute cannot be found, the program tries to fallback
          to the base state of the target state, which is specified with `"base": "<State name>"`.
        - If still fails, fallback to the rest state.
        - If still, which means that the attribute is not specified in the current theme, the
          program will try the parent theme, then the parent of parent theme, and repeat this until
          it reaches the default theme (or more accurately, the root theme).
        - As the root theme must contain all attributes available, if the selector or attribute
          still cannot be found even in root theme, the program throws an error.

        :param selector: The selector to the style, or a parsed list
        :param attr_name: The attribute name
        :param allow_not_found: True to return False when attribute not found, otherwise raise
                                error in such cases
        :return: The attribute _value
        """
        parsed_selector = selector if type(selector) is list else self.parse_selector(selector)
        style = self.select(parsed_selector, copy=False, allow_not_found=False)
        if attr_name not in style:
            # Fallback mechanism
            # Fallback to base or rest state and try
            if parsed_selector[-1] != "rest":
                new_selector = parsed_selector
                new_selector[-1] = "rest" if "base" not in style else style["base"]
                return self.get_style_attr(new_selector, attr_name)
            # If still fails, fallback to parent
            if self.parent is not None:
                return self.parent.get_style_attr(selector, attr_name)
            elif self.parent is None and self.name != CTheme.DEFAULT_THEME.name:
                return CTheme.DEFAULT_THEME.get_style_attr(selector, attr_name)
            # If is already default theme (no parent), then go fuck your selector
            if self.name == CTheme.DEFAULT_THEME.name:
                if allow_not_found:
                    return False
                else:
                    raise CStyleNotFoundError(
                        f"Attribute <{attr_name}> is not exsited in the default theme. "
                        "Check your selector!"
                    )
        if type(style[attr_name]) is dict:
            return style[attr_name].copy()
        else:
            return style[attr_name]

    def get_preset_color(self, color_name: str):
        """Find a preset color from color palette.

        Example
        -------
        .. code-block:: python
            my_theme = CTheme()
            white = my_theme.get_preset_color("-white")
            default_bg = my_theme.get_preset_color("default_bg")
        This shows getting the white color and the preset color named `default_bg`.

        :param color_name: Name of the color
        """
        keywords = {
            "-transparent": (0, 0, 0, 0),
            "-black": (0, 0, 0, 255),
            "-white": (255, 255, 255, 255),
            "-absneutralgrey": (128, 128, 128, 255),
            "-errcolor": color.ERR_COLOR,
        }
        if color_name in keywords:
            # If is a keyword
            result = keywords[color_name]
        else:
            # If not a keyword

            if color_name in self.color_palette:
                # If existed
                result = self.color_palette[color_name]
            else:
                # If not, fallback
                if isinstance(self.parent, CTheme):
                    # If it has parent, fallback to parent
                    result = self.parent.get_preset_color(color_name)
                else:
                    # If not, then is root theme, go fuck your color name
                    warnings.warn(
                        f"Color <{color_name}> if not found anywhere.",
                        color.CColorWarning,
                    )
                    result = color.ERR_COLOR
        if isinstance(result, dict):
            return result.copy()
        else:
            return result

    def mixin(self, selector: str, new_style: dict, copy: bool = False) -> "CTheme":
        """Mix, or in other words, override custom styles into the theme.

        Example
        -------
        .. code-block:: python
            my_theme = CTheme()
            my_theme.mixin("CButton.ITSELF", {"rest": {"background": (255, 0, 0, 255)},
                                               "hover": {"background": (0, 0, 255, 255)}})
            my_subtheme = my_theme.mixin("CButton:hover", {"background": (255, 0, 0, 255)},
                                         copy=True)
        The first line shows mixing in a red background style at rest state and a blue background
        style at hover state into CButton. The second line shows creating a subtheme base on
        `my_theme`, but with red background for `CButton` at `hover` state.

        :param selector: The selector string, indicates where to mix in
        :param new_style: A styles json, to be mixed in
        :param copy: Whether to copy a new theme, otherwise modify the current object
        :return theme_operate: The edited theme object, either self of copy of self
        """
        if copy:
            theme_operate = CTheme(self.styles)
        else:
            theme_operate = self
        style_operate = theme_operate.select(selector, copy=False, allow_not_found=False)
        style_operate.update(new_style)
        return theme_operate

    def special(self, selector: str, **kwargs) -> "CTheme":
        """Create a sub-theme with few modifications on the theme.

        Can be used when applying custom styles on a specific widget.

        Example
        -------
        .. code-block:: python
            CButton(window, styles=my_theme.special(background=(255, 0, 0, 255)))
        This shows setting a `CButton`'s style base on `my_theme`, but with background red.

        :param selector: The selector string, indicates where to mix in
        :param kwargs: Styles to change
        :return new_theme: The modified CTheme object
        """
        ## Handling <CWidget.ITSELF> selectors
        if "ITSELF" in selector:
            # special() does not support CWidget.ITSELF, as it simply changes attributes.
            warnings.warn(
                "<CWidget.ITSELF> is not supported by CTheme.special()! "
                "It will be regarded as <CWidget.rest>"
            )
            # So we just simply regard any <ITSELF> as <rest>, as in this case the user may want to
            # change the default appearance.
            selector = selector.replace("ITSELF", "rest")
        ## Creating a modified sub-theme
        # Create a new theme with required stuff
        # (This is to prevent modifications on original theme)
        new_theme = CTheme({}, parent=self)
        new_theme.is_special = True
        # Modifying styles of the new theme
        style_operate = new_theme.select(selector, copy=False, make_path=True)
        style_operate.update(kwargs)
        # Renaming the new theme
        existed_special_count = 0
        for child in self.children:
            if child.is_special:
                existed_special_count += 1
        new_theme.rename(
            f".special{existed_special_count}",
            f"{self.friendly_name} (Special {existed_special_count})",
        )
        # (Shall we delete unnecessary data from the modified theme in the future?)
        ## Returning the modified theme
        return new_theme

    def apply_on(self, widget) -> CTheme:
        """Apply theme on a widget.

        Example
        -------
        .. code-block:: python
            my_theme = CTheme()
            my_button = CButton(my_window, text="Hello world")
            my_theme.apply_on(my_button)
        This shows applying theme on a `CButton`

        :param widget: The widget to apply theme to
        :return self: The theme itself
        """
        widget.apply_theme(self)
        return self


# Load internal themes
CTheme._load_internal_themes()

# Alias for default theme
light_theme = default_theme = CTheme.DEFAULT_THEME
dark_theme = CTheme.INTERNAL_THEMES["default.dark"]
sv_theme = sv_light_theme = CTheme.INTERNAL_THEMES["sun_valley.light"]
sv_dark_theme = CTheme.INTERNAL_THEMES["sun_valley.dark"]



"""Translated from original Chinese version / 译自中文原文:

python项目有两个极端：
一种是诗，一种是屎。
写诗的人可能看不起写屎的，
但是实际上也只是样貌不同罢了。
诗虽然整洁、美丽，但是读起来很费劲，没有注释那也是赤石。
屎就更离谱了，全是堆叠出来的，结构复杂的，能跑的，
但是我没说是谁能跑，总之写的时候是爽的。
两者都是写完过一会儿后没人能读懂的，或者是很费劲才能读懂的，
能直接读懂的我觉得已经飞升了。
如果你知道自己在做什么，
那就自己管理吧。

能跑就行了，我说我

—— Original text by LittleWhiteCloud / 中文原文作者：小白云"""

THE_POEM = """
There are two extremes in Python, 
one is art while the other is fart. 
Those who create art look down on those who fart, 
while they are not actually far apart. 

Art is filled with beauty making it murky. 
Art is just fart with comments and space bars. 
Fart is whatever piled up from nothing, 
messy but runs fast, 
except it is the programmer who runs away fast, 
while anyway (s)he types fast. 

Both art and fart aren't easy to parse; 
those who understand are a rare class. 

For those who want to make arts with arts... 
...or maybe fart: 
If you know what you are doing, 
then do it yourself. 

Never mind, 
it will be OK if it runs fast... 
I mean myself...

—— Translated by rgzz666 from original text by LittleWhiteCloud
"""


print(THE_POEM)


# The Charmy development team / Charmy 开发团队
# Yours truely / 敬上
#
# 2026/02/05 (Y/M/D)



from .object import CharmyObject


class Pos(CharmyObject):
    """Pos is a class to store position."""

    def __init__(self, x=0, y=0):
        super().__init__()
        self.new("x", x)
        self.new("y", y)

    def __call__(self, x: int | float | None = None, y: int | float | None = None):
        if x:
            self.set("x", x)
        if y:
            self.set("y", y)

    def __str__(self):
        """Return position in string."""
        return f"Pos({self.get('x')}, {self.get('y')})"



from ..const import MANAGER_ID, Backends
from ..object import CharmyObject
from ..widgets.cmm import CharmyManager


class Color(CharmyObject):
    """Color manager"""

    def __init__(self, *args, draw_framework: Backends = None, **kwargs):
        super().__init__(*args, **kwargs)

        # The Drawing Framework
        if draw_framework is None:
            # Auto find CharmyManager Object
            self.manager: CharmyManager = self.get_obj(MANAGER_ID)
            self.new(
                "drawing.framework",
                self._get_drawing_framework(),
                get_func=self._get_drawing_framework,
            )
        else:
            self.new("drawing.framework", draw_framework)

        # Import Drawing Framework
        match self["drawing.framework"]:
            case Backends.SKIA:
                self.skia = self.manager.skia
            case _:
                raise ValueError("Not supported drawing framework")
        self.new("color_object", None)

    def set_color_rgba(
        self, r: int | float, g: int | float, b: int | float, a: int | float = 255
    ) -> None:
        """set_rgba(r=255, g=255, b=255, a=255) or set_rgba(r=1.0, g=1.0, b=1.0, a=1.0)

        Args:
            r (int | float): The red component of the color.
            g (int | float): The green component of the color.
            b (int | float): The blue component of the color.
            a (int | float): The alpha component of the color.

        Returns:
            None
        """
        match self["drawing.framework"]:
            case Backends.SKIA:
                self["color_object"] = self.skia.Color(
                    self._c(r), self._c(g), self._c(b), self._c(a)
                )

    def set_color_hex(self, _hex: str) -> None:
        """
        Convert hex color string to color.

        Args:
            _hex (str): Hex color string (support #RRGGBB and #RRGGBBAA format)

        Returns:
            skia.Color: Corresponding RGBA color object

        Raises:
            ValueError: When hex color format is invalid
        """
        hex_color = _hex.lstrip("#")
        if len(hex_color) == 6:  # RGB 格式，默认不透明(Alpha=255)
            r = int(hex_color[0:2], 16)
            g = int(hex_color[2:4], 16)
            b = int(hex_color[4:6], 16)
            match self["drawing.framework"]:
                case Backends.SKIA:
                    self["color_object"] = self.skia.ColorSetRGB(r, g, b)  # 返回不透明颜色
        elif len(hex_color) == 8:  # RGBA 格式(含 Alpha 通道)
            r = int(hex_color[0:2], 16)
            g = int(hex_color[2:4], 16)
            b = int(hex_color[4:6], 16)
            a = int(hex_color[6:8], 16)
            match self["drawing.framework"]:
                case Backends.SKIA:
                    self["color_object"] = self.skia.ColorSetARGB(a, r, g, b)  # 返回含透明度的颜色
        else:
            raise ValueError("HEX Should be #RRGGBB or #RRGGBBAA format")

    def set_color_name(self, name: str) -> None:
        """Convert color name string to color.

        Args:
            name (str): Color name
        Returns:
            skia.Color: Corresponding RGBA color object
        Raises:
            ValueError: When color not exists
        """
        match self["drawing.framework"]:
            case Backends.SKIA:
                try:
                    self["color_object"] = getattr(self.skia, f"Color{name.upper()}")
                except:
                    raise ValueError(f"Unknown color name: {name}")

    @staticmethod
    def _c(x):
        """Convert float color component to int.

        Args:
            x (int | float): Color component value (0.0-1.0 or 0-255)

        Returns:
            int: Corresponding int value (0-255)
        """
        if isinstance(x, float):
            if 0 < x <= 1.0:
                x = x * 255
        return x

    def _get_drawing_framework(self):
        return self.manager["drawing.framework"]



from __future__ import annotations as _

import collections.abc
import re
import threading
import time
import typing
import warnings

from .const import ID
from .object import CharmyObject


class EventHandling(CharmyObject):

    # fmt: off
    EVENT_TYPES: list[str] = [
        # Window size & pos
        "resize", "move", 
        # Widget internal changes
        "configure", "update", "draw",
        # Mouse events
        "mouse_move", "mouse_enter", "mouse_leave", "mouse_press", "mouse_release", "click", 
        "double_click",
        # Focus events
        "focus_gain", "focus_loss", 
        # Key events
        "key_press", "key_release", "key_repeat", "char", 
        # Special events
        "delay", "repeat", 
    ]
    # fmt: on
    multithread_tasks: list[tuple[EventTask, Event]] = []
    WORKING_THREAD: threading.Thread

    @staticmethod
    def _execute_task(task: EventTask | DelayTask, event_obj: Event) -> None:
        """To execute the bound task directly, regardless its props, mainly for internal use."""
        match task.target:
            case _ if callable(task.target):
                task.target(event_obj)
            case _ if isinstance(task.target, collections.abc.Iterable):
                for task_step in task.target:
                    task_step(event_obj)
            case _:
                raise ValueError(
                    "Error type for suzaku Task target! Excepted callable or "
                    f"iterable but received {type(task.target)}"
                )

    def __init__(self):
        """A class containing event handling abilities.

        Example
        -------
        This is mostly for internal use of Charmy.

        .. code-block:: python

            class Widget(EventHandling, ...):
                def __init__(self):
                    super().__init__(self)
            ...

        This shows subclassing EventHandling to let Widget gain the ability of handling events.
        """
        super().__init__()
        self.latest_event: Event = Event(widget=None, event_type="NO_EVENT")
        self.tasks: dict[str, list[EventTask]] = {}
        ## Initialize tasks list
        for event_type in self.__class__.EVENT_TYPES:
            self.tasks[event_type] = []

    def parse_event_type_str(self, event_type_str: str) -> dict:  # NOQA
        """This function parses event type string.

        :param event_type_str: The event type string to be parsed
        :returns: JSON, parsed event type
        """
        if not re.match(".*\\[.*\\]", event_type_str):  # NOQA
            return {"type": event_type_str, "params": []}
        event_type = re.findall("^(.*?)\\[", event_type_str)[0]
        params_raw = re.findall("\\[(.*?)\\]$", event_type_str)[0]  # NOQA
        params = params_raw.split(",")
        if len(params) == 1:
            if params[0].strip() == "":
                params = []
        return {"type": event_type, "params": params}

    def execute_task(self, task: EventTask, event_obj: Event | None = None):
        """To execute a task

        Example
        -------
        .. code-block:: python

            my_task = Widget.bind("delay[5]", lambda: print("Hello Suzaku"))
            Widget.execute_task(my_task)

        """
        if event_obj is None:
            event_obj = Event()
        assert event_obj is not None
        if event_obj.widget is None:
            event_obj.widget = self
        if not task.multithread:
            # If not multitask, execute directly
            EventHandling._execute_task(task, event_obj)
            # If is a delay event, it should be removed right after execution
            if isinstance(task, DelayTask):
                self.unbind(task)
        else:
            # Otherwise add to multithread tasks list and let the working thread to deal with it
            # If is a delay task, should add some code to let it unbind itself, here is a way,
            # which is absolutely not perfect, though works, to implement this mechanism, by
            # overriding its target with a modified version
            def self_destruct_template(task, event_obj):
                EventHandling._execute_task(task, event_obj)
                self.unbind(task)

            if isinstance(task, DelayTask):
                task.target = lambda event_obj: self_destruct_template(task, event_obj)
            EventHandling.multithread_tasks.append((task, event_obj))

    def trigger_event(self, event_obj: Event) -> None:
        """To trigger a type of event

        Args:
            event_type: The type of event to trigger

        Example
        -------
        .. code-block:: python

            class Widget(EventHandling, ...):
                ...

            my_widget = Widget()
            my_widget.trigger("mouse_press")

        This shows triggering a `mouse_press` event in a `Widget`, which inherited
        `EventHandling` so has the ability to handle events.
        """
        # Parse event type string
        parsed_event_type = self.parse_event_type_str(event_obj.event_type)
        # Create a default Event object if not specified
        if event_obj is None:
            event_obj = Event(widget=self, event_type=tuple(parsed_event_type.keys())[0])
        # Add the event to event lists (the widget itself and the global list)
        self.latest_event = event_obj
        Event.latest = event_obj
        # Find targets
        targets = [parsed_event_type["type"]]
        if not parsed_event_type["params"]:
            targets.append(parsed_event_type["type"] + "[*]")
        else:
            targets.append(event_obj.event_type)
        # if parsed_event_type["params"][0] in ["", "*"]: # If match all
        #     targets.append(parsed_event_type["type"])
        #     targets.append(parsed_event_type["type"] + "[*]")
        for target in targets:
            if target in self.tasks:
                for task in self.tasks[target]:
                    # To execute all tasks bound under this event
                    self.execute_task(task, event_obj)

    trigger = trigger_event  # Alias for trigger

    def bind(
        self,
        event_type: str,
        target: typing.Callable | typing.Iterable,
        multithread: bool = False,
        _keep_at_clear: bool = False,
    ) -> EventTask | bool:
        """To bind a task to the object when a specific type of event is triggered.

        Example
        -------
        .. code-block:: python

            my_button = Button(...).pack()
            press_down_event = my_button.bind("mouse_press", lambda _: print("Hello world!"))

        This shows binding a hello world to the button when it's press.

        :param event_type: The type of event to be bound to
        :param target: A (list of) callable thing, what to do when this task is executed
        :param multithread: If this task should be executed in another thread (False by default)
        :param _keep_at_clear: If the task should be kept when cleaning the event's binding
        :return: EventTask that is bound to the task if success, otherwise False
        """
        parsed_event_type = self.parse_event_type_str(event_type)
        if parsed_event_type["type"] not in self.__class__.EVENT_TYPES:
            # warnings.warn(f"Event type {event_type} is not present in {self.__class__.__name__}, "
            #                "so the task cannot be bound as expected.")
            # return False
            self.EVENT_TYPES.append(event_type)
        if event_type not in self.tasks:
            self.tasks[event_type] = []
        task_id = f"{self.id}.{event_type}.{len(self.tasks[event_type])}"
        # e.g. CButton114.focus_gain.514 / CEventHandling114.focus_gain.514
        match parsed_event_type["type"]:
            case "delay":
                raise NotImplementedError("Delay tasks are not implemented yet!")
                task = DelayTask(  # NOQA
                    target,  # I will fix this type error later (ignore is ur type check is off)
                    parsed_event_type["params"][0],
                    multithread,
                    _keep_at_clear,
                    task_id,  # NOQA
                )
            case "repeat":
                raise NotImplementedError("Repeat tasks is not implemented yet!")
            case _:  # All normal event types
                task = EventTask(target, multithread, _keep_at_clear, task_id)
        self.tasks[event_type].append(task)
        return task

    def find_task(self, task_id: str) -> EventTask | bool:
        """To find an event task using task ID.

        Example
        -------

        .. code-block:: python

            my_button = Button(...)
            press_task = my_button.find_task("Button114.mouse_press.514")

        This shows getting the `EventTask` object of task with ID `Button114.mouse_press.514`
        from bound tasks of `my_button`.

        :return: The EventTask object of the task, or False if not found
        """
        task_id_parsed = task_id.split(".")
        if len(task_id_parsed) == 2:  # If is a shortened ID (without widget indicator)
            task_id_parsed.insert(0, self.id)  # We assume that this indicates self
        for task in self.tasks[task_id_parsed[1]]:
            if task.id == task_id:
                return task
        else:
            return False

    def unbind(self, target_task: str | EventTask | DelayTask) -> bool:
        """To unbind the task with specified task ID.

        Example
        -------

        .. code-block:: python

            my_button = Button(...)
            my_button.unbind("Button114.mouse_press.514")

        This show unbinding the task with ID `Button114.mouse_press.514` from `my_button`.

        .. code-block:: python

            my_button = Button(...)
            my_button.unbind("Button114.mouse_press.*")
            my_button.unbind("mouse_release.*")

        This show unbinding all tasks under `mouse_press` and `mouse_release` event from
        `my_button`.

        :param target_task: The task ID or `EventTask` to unbind.
        :return: If success
        """
        match target_task:
            case str():  # If given an ID string
                for task_index, task in enumerate(self.tasks[target_task]):
                    if task.id == target_task:
                        self.tasks[target_task].pop(task_index)
                        return True
                else:
                    return False
            case EventTask():
                for event_type in self.tasks:
                    if target_task in self.tasks[event_type]:
                        self.tasks[event_type].remove(target_task)
                        return True
                else:
                    return False
            case _:
                warnings.warn(
                    "Wrong type for unbind()! Must be event ID or task object",
                    UserWarning,
                )
                return False


class Event(CharmyObject):
    """Used to represent an event."""

    latest: "Event"

    def __init__(  # NOQA
        self,
        widget: EventHandling | None = None,
        event_type: str = "[Unspecified]",
        **kwargs,
    ):
        """This class is used to represent events.

        Some properties owned by all types of events are stored as attributes, such as widget and type.
        Others are stored as items, which can be accessed or manipulated just like dict, e.g.
        `CEvent["x"]` for get and `CEvent["y"] = 16` for set.

        Included in description.

        :param widget: The widget of the event, None by default
        :param event_type: Type of the event, in string, `"[Unspecified]"` by default
        :param **kwargs: Other properties of the event, will be added as items
        """
        self.event_type: str = event_type  # Type of event
        self.widget: typing.Optional[typing.Any] = widget  # Relating widget
        self.window_base: typing.Optional[typing.Any] = None  # WindowBase of the current window
        self.window: typing.Optional[typing.Any] = None  # Current window
        self.event_data: dict = {}
        # Not all properties above will be used
        # Update stuff from args into attributes
        for prop in kwargs.keys():
            if prop not in ["widget", "event_type"]:
                self[prop] = kwargs[prop]

    def __setitem__(self, key: str, value: typing.Any):
        self.event_data[key] = value

    def __getitem__(self, key: str) -> typing.Any:
        if key in self.event_data:
            return self.event_data[key]
        else:
            return None  # If no such item avail, returns None


Event.latest = Event(widget=None, event_type="NO_EVENT")


class EventTask:
    """A class to represent event task when an event is triggered."""

    def __init__(
        self,
        target: typing.Callable | typing.Iterable,
        multithread: bool = False,
        _keep_at_clear: bool = False,
        id_: str | int | typing.Literal[ID.AUTO] = ID.AUTO,
    ):
        """Each object is to represent a task bound to the event.

        Example
        -------
        This is mostly for internal use of suzaku.

        .. code-block:: python

            class EventHandling():
                def bind(self, ...):
                    ...
                    task = EventTask(event_id, target, multithread, _keep_at_clear)
                    ...

        This shows where this class is used for storing task properties in most cases.

        :param target: A callable thing, what to do when this task is executed
        :param multithread: If this task should be executed in another thread (False by default)
        :param _keep_at_clear: If the task should be kept when cleaning the event's binding
        :param id_: The task id of this task
        """
        self.id: str | int | typing.Literal[ID.AUTO] = id_
        self.target: typing.Callable | typing.Iterable = target
        self.multithread: bool = multithread
        self.keep_at_clear: bool = _keep_at_clear


class DelayTask(EventTask):
    NotImplemented  # NOQA


class WorkingThread(threading.Thread, CharmyObject):
    """CWorkingThread is a class represent the event working thread."""

    def __init__(self, *args, **kwargs):
        threading.Thread.__init__(self, *args, **kwargs)
        CharmyObject.__init__(self, id_="event.main_thread")
        self.tasks: list[tuple[Event, EventTask | DelayTask]] = []  # [(event, task), ...]
        self.is_alive: bool = True
        self.lock: threading.Lock = threading.Lock()

    def execute_tasks(self):
        self.lock.acquire()
        for task in self.tasks:
            if callable(task[1].target):
                task[1].target(task[0])
            elif isinstance(task[1], collections.abc.Iterable):
                # Function list
                for func in task[1]:
                    func(task[0])
            self.tasks.remove(task)
        else:
            # warnings.warn("")
            pass
        self.lock.release()

    def run(self):
        while self.is_alive:
            self.execute_tasks()
            if len(self.tasks) == 0:
                # If idle, then rest for 0.02s to save CPU time
                time.sleep(0.02)

    def add_task(self, task: EventTask | DelayTask, event: Event):
        self.tasks.append((event, task))
